#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Path traversal manipulation."""

from collections import deque
import copy
from itertools import chain

from .config import Config
from .errors import SelectiveNoPathError


def _get_all_subflows_dict(flow_name):
    """Get all subflows for the given flow name, the result is stored in a dict.

    The resulting dict has keys that correspond to all transitive subflows for flow flow_name and values for a subflow
    is a list of flows that are directly parent for corresponding subflow (key).

    E.g.:
      flow1:
        subflows: flow2, flow3

      flow2:
        subflows: flow1, flow3, flow4

    The resulting dict for _get_all_subflows_dict('flow1') will be:
       {
         "flow2": ["flow1"],
         "flow3": ["flow1", "flow2"],
         "flow4": ["flow2"],
         "flow1": ["flow2"],
       }

    Read as: I can get directly to flow <key> from flows <values>.


    :param flow_name: a flow name for which all subflows should be computed
    :return: dict representing all subflows and direct paths for the given subflow
    """
    stack = deque()
    result = {}

    stack.append(flow_name)
    while stack:
        inspected_flow_name = stack.pop()
        for edge in Config.edge_table[inspected_flow_name]:
            for node_name in edge['to']:
                if Config.is_flow(node_name):
                    if node_name not in result:
                        result[node_name] = set()
                        stack.append(node_name)
                    result[node_name] |= {inspected_flow_name}

    for key, value in result.items():
        result[key] = list(value)

    return result


def _normalize_path(paths):
    """Normalize multiple graph traversals by edges into one that traverses all edges.

    In general we can get to a task by multiple paths. As we would like to ensure that all paths are traversed, we
    compound multiple traversals into a single one that traverses all necessary edges.

    :param paths: a list of paths that should be compound into a single traversal
    :return: a dict representing compound traversal
    """
    result = {}
    for entry in paths:
        for key, value in entry.items():
            if key not in result:
                result[key] = []
            result[key] = list(set(result[key]) | set(value))

    return result


def _compute_paths(flow_name, task_name):
    """Compute all paths in a flow to a node.

    :param flow_name: name of flow that should be traversed
    :param task_name: a name of node that should be visited
    :return: a list of paths that lead to the node that should be visited
    """
    result = []
    stack = deque()
    stack.append(({task_name}, set(), {}))

    while stack:
        to_expand, visited, expanded = stack.pop()

        for name in to_expand:
            for edge_idx, edge in enumerate(Config.edge_table[flow_name]):
                if name in edge['to'] and (edge_idx not in expanded.keys() or name not in expanded[edge_idx]):
                    new_expanded = copy.copy(expanded)
                    if edge_idx not in new_expanded:
                        new_expanded[edge_idx] = [name]
                    else:
                        new_expanded[edge_idx].append(name)

                    new_visited = visited | {name}
                    new_to_expand = to_expand | set(edge['from']) - {name}

                    if new_visited == new_to_expand:
                        result.append(new_expanded)
                    else:
                        stack.append((new_to_expand, new_visited, new_expanded))

    return _normalize_path(result)


def _raise_for_result_check(task_names, path):
    """Check that all nodes stated in task_names are present in path - i.e. they are visited during traversing.

    :param task_names: a list of node names that should be checked
    :param path: traverses that are per-flow specific
    :raises SelectiveNoPathError: there was not found any path to the given node
    """
    visited_marking = dict.fromkeys(task_names, False)
    for node in visited_marking:
        for flow in path:
            if any(n == node for n in chain(*path[flow].values())):
                visited_marking[node] = True

    for node, visited in visited_marking.items():
        if not visited:
            raise SelectiveNoPathError("No path to node '%s' found" % node)


def _compute_subsequent_edges(flow_name, node_names):
    """Compute nodes that are subsequent nodes based on node_names.

    :param flow_name: name of the flow in which subsequent nodes should be found
    :param node_names: a list of nodes that were run, note that they does not need to be necessarily stated in flow_name
    :return: a list of tasks that follow after node_names execution
    """
    result = []
    desired_nodes = set(node_names)

    change = True
    while change:
        change = False
        for edge_idx, edge in enumerate(Config.edge_table[flow_name]):
            if edge['from'] and edge_idx not in result and set(edge['from']).issubset(desired_nodes):
                result.append(edge_idx)
                desired_nodes |= set(edge['to'])
                change = True

    return result


def _compute_traversals(flow_name, task_names, follow_subflows=True):
    """Compute all traversals/paths to nodes from a flow.

    :param flow_name: a name of flow to start traversing with
    :param task_names: a list of nodes we want to visit/traverse
    :param follow_subflows: if True, we also inspect transitively all subflows from flow_name
    :return: a dict, where keys are flows that need to be traversed and values are traversals/paths to all task_names
    """
    result = {}
    stack = deque()
    subflows_dict = {}
    traversed_subflows = set()

    if not isinstance(task_names, (list, tuple)):
        task_names = [task_names]

    for task_name in task_names:
        stack.append((flow_name, task_name))

    if follow_subflows:
        subflows_dict = _get_all_subflows_dict(flow_name)
        for subflow_name in subflows_dict.keys():  # pylint: disable=consider-iterating-dictionary
            for task_name in task_names:
                stack.append((subflow_name, task_name))

    while stack:
        flow, node = stack.pop()
        paths = _compute_paths(flow, node)

        if flow != flow_name and paths:
            for parent_flow in subflows_dict[flow]:
                if (parent_flow, flow) not in traversed_subflows:
                    traversed_subflows |= {(parent_flow, flow)}
                    stack.append((parent_flow, flow))

        if flow in result:
            result[flow] = _normalize_path((result[flow], paths))
        else:
            result[flow] = paths

    _raise_for_result_check(task_names, result)
    return result


def compute_selective_run(flow_name, task_names, follow_subflows=False, run_subsequent=False):
    """Compute selective run for a flow.

    :param flow_name: a name of the flow that should be run
    :param task_names: a list of tasks that should be run
    :param follow_subflows: apply selective run to all subflows (transitively)
    :param run_subsequent: run tasks that depend on task_names
    :return: computed selective run dictionary
    """
    traversals = _compute_traversals(flow_name, task_names, follow_subflows)
    result = {
        'task_names': task_names,
        'waiting_edges_subset': traversals
    }

    if run_subsequent:
        if isinstance(run_subsequent, (list, tuple)):
            subsequent_flows = run_subsequent
        else:
            subsequent_flows = traversals.keys()

        for flow in subsequent_flows:
            subsequent_edges = _compute_subsequent_edges(flow, task_names)
            for idx in subsequent_edges:
                # We need to make sure that we start all nodes so we have fire edge when we visit it twice due to cycles
                #
                #      T1 <-
                #    / |    |
                #   /  |    |
                # T3    T2 --
                #
                # T2 should be run, if we get minimal path to T2, there will be no T3, but since we have edge T2->T1, we
                # need to start T3 as subsequent.
                #
                traversals[flow][idx] = Config.edge_table[flow][idx]['to']

    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Selinon flow scheduling (entrypoint) for a user."""

import logging

from .config import Config
from .dispatcher import Dispatcher
from .errors import UnknownFlowError
from .selective import compute_selective_run
from .trace import Trace

_logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _do_run_flow(kwargs, queue):
    """Actually run the flow, return dispatcher id."""
    dispatcher_id = Dispatcher().apply_async(kwargs=kwargs, queue=queue)
    trace_msg = {
        'flow_name': kwargs["flow_name"],
        'node_args': kwargs["node_args"],
        'dispatcher_id': str(dispatcher_id),
        'queue': queue,
        'selective': kwargs.get("selective"),
    }
    Trace.log(Trace.FLOW_SCHEDULE, trace_msg)
    return dispatcher_id


def run_flow(flow_name, node_args=None):
    """Run flow by it's name.

    :param flow_name: name of the flow to be run
    :param node_args: arguments that will be supplied to flow
    :return: flow ID (dispatcher ID)
    """
    if Config.dispatcher_queues is None or flow_name not in Config.dispatcher_queues:
        raise UnknownFlowError("No flow with name '%s' defined" % flow_name)

    queue = Config.dispatcher_queues[flow_name]
    _logger.debug("Scheduling flow '%s' with node_args '%s' on queue '%s'", flow_name, node_args, queue)
    kwargs = {'flow_name': flow_name, 'node_args': node_args}
    return _do_run_flow(kwargs, queue)


def run_flow_selective(flow_name, task_names, node_args, follow_subflows=False, run_subsequent=False):
    """Run only desired tasks in a flow.

    :param flow_name: name of the flow that should be run
    :param task_names: name of the tasks that should be run
    :param node_args: arguments that should be supplied to flow
    :param follow_subflows: if True, subflows will be followed and checked for nodes to be run
    :param run_subsequent: trigger run of all tasks that depend on the desired task
    :return: dispatcher id that is scheduled to run desired selective task flow
    :raises selinon.errors.SelectiveNoPathError: there was no way found to the desired task in the flow
    """
    selective = compute_selective_run(flow_name, task_names, follow_subflows, run_subsequent)
    queue = Config.dispatcher_queues[flow_name]

    _logger.debug("Scheduling selective flow '%s' with node_args '%s' on queue '%s', computed selective run state: %s",
                  flow_name, node_args, queue, selective)
    kwargs = {'flow_name': flow_name, 'node_args': node_args, 'selective': selective}
    return _do_run_flow(kwargs, queue)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Predicate interface - predicate for building conditions."""

import abc

import codegen

from .errors import ConfigurationError
from .helpers import check_conf_keys
from .helpers import dict2json


class Predicate(metaclass=abc.ABCMeta):
    """An abstract predicate representation."""

    @abc.abstractmethod
    def __init__(self):
        """Instantiate predicate representation."""

    @abc.abstractmethod
    def __str__(self):
        """Create a string representation of this predicate.

        :return: string representation (Python code)
        :rtype: str
        """

    @abc.abstractclassmethod
    def create(cls, tree, nodes_from, flow, can_inspect_results):
        """Create the predicate.

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param nodes_from: nodes which are used within edge definition
        :type nodes_from: List[Nodes]
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :param can_inspect_results: True if predicates in the condition can query task result
        :type can_inspect_results: bool
        :return: Predicate instance
        """

    @abc.abstractmethod
    def ast(self):
        """Python AST of this predicate (construct transitively for all indirect children as well).

        :return: AST of describing all children predicates
        """

    @abc.abstractmethod
    def predicates_used(self):
        """Compute all predicates that are used (transitively) by child/children.

        :return: used predicates by children
        :rtype: List[Predicate]
        """

    @abc.abstractmethod
    def nodes_used(self):
        """Compute all nodes that are used (transitively) by child/children.

        :return: list of nodes that are used
        :rtype: List[Node]
        """

    @staticmethod
    def construct_default(flow):
        """Construct default predicate for edge.

        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :rtype: Predicate
        """
        from .builtin_predicate import AlwaysTruePredicate

        return AlwaysTruePredicate(flow=flow)

    @staticmethod
    def construct_default_dict():
        """Construct default predicate definition as it would be written in the configuration file."""
        return {'name': 'alwaysTrue'}

    @staticmethod
    def construct(tree, nodes_from, flow, can_inspect_results=True):  # pylint: disable=too-many-branches
        """Top-down creation of predicates - recursively called to construct predicates.

        :param tree: a dictionary describing nodes
        :type tree: dict
        :param nodes_from: nodes which are used within edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :param can_inspect_results: True if predicates in the condition can query task result
        :type can_inspect_results: bool
        :rtype: Predicate
        """
        from .leaf_predicate import LeafPredicate
        from .builtin_predicate import OrPredicate, AndPredicate, NotPredicate

        if not tree:
            raise ConfigurationError("Bad condition '%s'" % tree)

        if 'name' in tree:
            if 'node' in tree:
                node = None
                for node_from in nodes_from:
                    if node_from.name == tree['node']:
                        node = node_from
                        break
                if node is None:
                    raise ConfigurationError("Node '%s' listed in predicate '%s' is not requested in 'nodes_from'"
                                             % (tree['node'], tree['name']))
            else:
                if len(nodes_from) == 1:
                    node = nodes_from[0]
                else:
                    # e.g. starting edge has no nodes_from
                    node = None

            unknown_conf = check_conf_keys(tree, known_conf_opts=('name', 'node', 'args'))
            if unknown_conf:
                raise ConfigurationError("Unknown configuration option for predicate '%s' in flow '%s': %s"
                                         % (tree['name'], flow.name, unknown_conf.keys()))

            predicate = LeafPredicate.create(tree['name'], node, flow, tree.get('args'))

            if not can_inspect_results and predicate.requires_message():
                raise ConfigurationError("Cannot inspect results of tasks '%s' in predicate '%s' in flow '%s'"
                                         % (nodes_from, tree['name'], flow.name))
            return predicate

        if 'or' in tree:
            return OrPredicate.create(tree['or'], nodes_from, flow, can_inspect_results)

        if 'not' in tree:
            return NotPredicate.create(tree['not'], nodes_from, flow, can_inspect_results)

        if 'and' in tree:
            return AndPredicate.create(tree['and'], nodes_from, flow, can_inspect_results)

        raise ConfigurationError("Unknown predicate:\n%s" % dict2json(tree))

    @staticmethod
    def construct_condition_name(flow_name, idx):
        """Create condition name for a dump.

        :param flow_name: flow name
        :type flow_name: str
        :param idx: index of condition within the flow
        :type idx: int
        :return: condition function representation
        """
        assert idx >= 0  # nosec
        return '_condition_{}_{}_cond'.format(flow_name, idx)

    def to_source(self):
        """Construct predicate source code.

        :return: predicate source code
        """
        return codegen.to_source(self.ast())

    @abc.abstractmethod
    def check(self):
        """Recursively/transitively check predicate correctness.

        :raises ValueError: if predicate is not correct
        """

    @abc.abstractmethod
    def requires_message(self):
        """Recursively check if one of the predicates require message from storage (result of previous task).

        :return: True if a result from storage is required
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import json
import pytest
from flexmock import flexmock
from selinon_test_case import SelinonTestCase
from selinon import Dispatcher
from selinon.system_state import SystemState
from selinon.errors import FlowError
from request_mock import RequestMock


class TestDispatcher(SelinonTestCase):
    def test_run(self):
        flow_name = 'flow1'
        edge_table = {
            flow_name: [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)
        state_dict = {'failed_nodes': {}, 'finished_nodes': {'Task1': ['<task-id>']}, 'active_nodes': []}

        flexmock(SystemState).should_receive('update').and_return(None)
        flexmock(SystemState).should_receive('to_dict').and_return(state_dict)

        dispatcher = Dispatcher()
        dispatcher.request = RequestMock()

        assert dispatcher.run(flow_name) == state_dict

    def test_dispacher_error(self):
        flow_name = 'flow1'
        edge_table = {
            flow_name: [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        exc = KeyError("some exception in dispatcher")
        flexmock(SystemState).should_receive('update').and_raise(exc)

        dispatcher = Dispatcher()
        dispatcher.request = RequestMock()
        flexmock(dispatcher).should_receive('retry').and_return(exc)

        # We should improve this by actually checking exception value
        with pytest.raises(KeyError):
            dispatcher.run(flow_name)

    def test_flow_error_retry(self):
        def my_retry(max_retries, exc):
            assert max_retries == 0
            assert json.loads(str(exc)) == json.loads(str(raised_exc))
            raise RuntimeError()  # we re-raise as stated in Celery doc

        flow_name = 'flow1'
        edge_table = {
            flow_name: [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)
        state_dict = {'finished_nodes': {}, 'failed_nodes': {}, 'active_nodes': []}

        raised_exc = FlowError(state_dict)
        flexmock(SystemState).should_receive('update').and_raise(raised_exc)  # we will retry
        flexmock(SystemState).should_receive('to_dict').and_return(state_dict)

        dispatcher = Dispatcher()
        dispatcher.request = RequestMock()
        flexmock(dispatcher).should_receive('retry').replace_with(my_retry)

        with pytest.raises(RuntimeError):
            dispatcher.run(flow_name)

    def test_retry(self):
        def my_retry(args, kwargs, countdown, queue):
            assert args == []
            assert countdown == 2
            assert queue == 'queue_flow1'
            state = kwargs.pop('state')
            assert set(state.keys()) == {
                'active_nodes',
                'failed_nodes',
                'finished_nodes',
                'waiting_edges',
                'triggered_edges'
            }
            assert len(state['active_nodes']) == 1
            assert set(state['active_nodes'][0].keys()) == {'id', 'name'}
            assert not state['failed_nodes']
            assert not state['finished_nodes']
            assert not state['waiting_edges']
            assert len(state['triggered_edges']) == 1
            assert kwargs == {
                'flow_name': 'flow1',
                'migration_version': 0,
                'node_args': None,
                'parent': None,
                'retried_count': 0,
                'retry': 2,
                'selective': False
            }
            # Celery will raise Celery's retry causing dispatcher to retry, but let's simulate
            # RuntimeError here for testing
            raise RuntimeError()

        flow_name = 'flow1'
        edge_table = {
            flow_name: [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        dispatcher = Dispatcher()
        dispatcher.request = RequestMock()
        flexmock(dispatcher).should_receive('retry').replace_with(my_retry)

        with pytest.raises(RuntimeError):
            dispatcher.run(flow_name)

    def test_selinon_retry(self):
        def my_retry(kwargs=None, max_retries=None, countdown=None, queue=None, exc=None):
            kwargs = kwargs or {}
            assert kwargs.get('flow_name') == flow_name
            assert kwargs.get('node_args') == node_args
            assert 'parent' in kwargs
            assert 'selective' in kwargs
            assert 'state' in kwargs
            assert max_retries is None
            assert countdown == 5
            assert queue == 'queue_flow1'
            assert exc is None
            raise RuntimeError()  # we re-raise as stated in Celery doc

        flow_name = 'flow1'
        node_args = {'foo': 'bar'}
        edge_table = {
            flow_name: [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table, retry_countdown={'flow1': 5}, max_retry={'flow1': 1})
        state_dict = {'finished_nodes': {'Task1': ['<task1-id>']}, 'failed_nodes': {}}

        flexmock(SystemState).should_receive('update').and_raise(FlowError(state_dict))

        dispatcher = Dispatcher()
        flexmock(dispatcher).should_receive('retry').replace_with(my_retry)
        dispatcher.request = RequestMock()

        with pytest.raises(RuntimeError):
            dispatcher.run(flow_name, node_args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Data storage interface."""

import abc


class SelinonMissingDataException(Exception):
    """Selinon exception that ressource was not found by storage"""
    pass

class DataStorage(metaclass=abc.ABCMeta):
    """Abstract Selinon storage adapter that is implemented by a user."""

    @abc.abstractmethod
    def __init__(self, *args, **kwargs):
        """Initialize storage.

        :param args: storage arguments as stated in YAML configuration file
        :param kwargs: storage key-value arguments as stated in YAML configuration (preferred over args)
        """

    @abc.abstractmethod
    def connect(self):
        """Connect to a resource, if not needed, should be empty."""
        raise NotImplementedError()

    @abc.abstractmethod
    def is_connected(self):
        """Check storage connection status.

        :return: True if connected to a resource
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def disconnect(self):
        """Disconnect from a resource."""
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve(self, flow_name, task_name, task_id):
        """Retrieve result stored in storage.

        :param flow_name: flow name in which task was executed
        :param task_name: task name that result is going to be retrieved
        :param task_id: id of the task that result is going to be retrieved
        :return: task result
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def store(self, node_args, flow_name, task_name, task_id, result):  # pylint: disable=too-many-arguments
        """Store result stored in storage.

        :param node_args: arguments that were passed to node
        :param flow_name: flow name in which task was executed
        :param task_name: task name that result is going to be stored
        :param task_id: id of the task that result is going to be stored
        :param result: result that should be stored
        :return: unique ID of stored record
        """
        raise NotImplementedError()

    def store_error(self, node_args, flow_name, task_name, task_id, exc_info):  # pylint: disable=too-many-arguments
        """Store information about task error.

        :param node_args: arguments that were passed to node
        :param flow_name: flow name in which task was executed
        :param task_name: task name that result is going to be stored
        :param task_id: id of the task that result is going to be stored
        :param exc_info: information about exception - tuple (type, value, traceback) as returned by sys.exc_info()
        :return: unique ID of stored record
        """
        # pylint: disable=abstract-method
        # no not mark this method with @abc.abstractmethod as we do not force a user to implement this
        raise NotImplementedError()

    def delete(self, flow_name, task_name, task_id):
        """Delete result stored in storage.

        :param flow_name: flow name in which task was executed
        :param task_name: task name that result is going to be retrieved
        :param task_id: id of the task that result is going to be retrieved
        """
        raise NotImplementedError("delete method is not implemented")

    def __del__(self):
        """Clean up."""
        if self.is_connected():
            self.disconnect()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Base class for user-defined tasks."""

import abc
import logging

from .celery import AsyncResult
from .errors import NoParentNodeError
from .errors import RequestError
from .errors import Retry
from .storage_pool import StoragePool


class SelinonTask(metaclass=abc.ABCMeta):
    """Base class for user-defined tasks."""

    def __init__(self, flow_name, task_name, parent, task_id, dispatcher_id):
        """Initialize SelinonTask (called by SelinonTaskEnvelope).

        :param flow_name: name of flow under which this tasks runs on
        :param task_name: name of task, note it can be aliased since we can have different task name and class name
        :param parent: direct task's predecessors stated in flow dependency
        :param task_id: id of this task
        :parent dispatcher_id: id of dispatcher handling the current flow
        """
        # pylint: disable=too-many-arguments
        self.flow_name = flow_name
        self.task_name = task_name
        self.parent = parent
        self.task_id = task_id
        self.dispatcher_id = dispatcher_id
        self.log = logging.getLogger(__name__)

    def _selinon_dereference_task_id(self, flow_names, task_name, index):
        """Compute task id based on mapping of ancestors (from parent sub-flows).

        :param flow_names: name of parent flow or list of flow names in case of nested flows
        :param task_name: name of task in parent flow
        :param index: index of result if more than one subflow was run
        :return: task id based from parent subflows
        """
        if not isinstance(flow_names, list):
            flow_names = [flow_names]

        parent_flow = self.parent
        for flow_name in flow_names:
            try:
                parent_flow = parent_flow[flow_name]
            except KeyError as exc:
                raise NoParentNodeError("No such parent flow '%s' for task '%s', check your configuration; nested "
                                        "as %s from flow %s"
                                        % (flow_name, self.task_name, flow_names, self.flow_name)) from exc
        try:
            task_id = parent_flow[task_name][index]
        except KeyError as exc:
            raise NoParentNodeError("No such parent task '%s' referenced by '%s' was run for task '%s' in flow '%s'"
                                    % (task_name, flow_names, self.task_name, self.flow_name)) from exc
        except IndexError as exc:
            raise NoParentNodeError("Requested index %s in parent task '%s' referencered by '%s', but there were "
                                    "run only %d tasks in task %s flow %s"
                                    % (index, task_name, flow_names, len(parent_flow[task_name]), self.task_name,
                                       self.flow_name)) from exc
        return task_id

    @property
    def storage(self):
        """Storage instance assigned to this task.

        :return: tasks's configured storage as stated in YAML config
        """
        return StoragePool.get_storage_by_task_name(self.task_name)

    def parent_task_result(self, parent_name):
        """Retrieve parent task result.

        :param parent_name: name of parent task to retrieve result from
        :return: result of parent task
        """
        try:
            parent_task_id = self.parent[parent_name]
        except KeyError as exc:
            raise NoParentNodeError("No such parent '%s' in task '%s' in flow '%s', check your configuration"
                                    % (parent_name, self.task_name, self.flow_name)) from exc

        return StoragePool.retrieve(self.flow_name, parent_name, parent_task_id)

    def parent_flow_result(self, flow_names, task_name, index=None):
        """Retrieve result of parent sub-flow task.

        Get parent subflow results; note that parent flows can return multiple results from task of same type
        because of loops in flows

        :param flow_names: name of parent flow or list of flow names in case of nested flows
        :param task_name: name of task in parent flow
        :param index: index of result if more than one subflow was run
        :return: result of task in parent subflow
        """
        index = -1 if index is None else index
        parent_flow_name = flow_names if not isinstance(flow_names, list) else flow_names[-1]
        task_id = self._selinon_dereference_task_id(flow_names, task_name, index)
        return StoragePool.retrieve(parent_flow_name, task_name, task_id)

    def parent_task_exception(self, parent_name):
        """Retrieve parent task exception. You have to call this from a fallback (direct or transitive).

        :param parent_name: name of task that failed (ancestor of calling task)
        :return: exception that was raised in the ancestor
        """
        try:
            parent_task_id = self.parent[parent_name]
        except KeyError as exc:
            raise NoParentNodeError("No such parent '%s' in task '%s' in flow '%s', check your configuration"
                                    % (parent_name, self.task_name, self.flow_name)) from exc

        # Celery stores exceptions in result field
        celery_result = AsyncResult(parent_task_id)
        if not celery_result.failed():
            raise RequestError("Parent task '%s' did not raised exception" % parent_name)

        return celery_result.result

    def parent_flow_exception(self, flow_names, task_name, index=None):
        """Retrieve parent task exception. You have to call this from a fallback (direct or transitive).

        :param flow_names: name of parent flow or list of flow names in case of nested flows
        :param task_name: name of task that failed (ancestor of calling task)
        :param index: index of result if more than one subflow was run
        :return: exception that was raised in the ancestor
        """
        index = -1 if index is None else index
        task_id = self._selinon_dereference_task_id(flow_names, task_name, index)

        # Celery stores exceptions in result field
        celery_result = AsyncResult(task_id)
        if not celery_result.failed():
            raise RequestError("Parent task '%s' from subflows %s did not raised exception"
                               % (task_name, str(flow_names)))

        return celery_result.result

    @staticmethod
    def retry(countdown=0):
        """Retry, always raises Retry, this is compatible with Celery's retry except you cannot modify arguments.

        :param countdown: countdown for rescheduling
        """
        raise Retry(countdown)

    @abc.abstractmethod
    def run(self, node_args):
        """Entrypoint - user defined computation.

        :param node_args: arguments passed to flow/node
        :return: tasks's result that will be stored in database as configured
        """



#!/usr/bin/env python3
"""Selinon - an advanced dynamic task flow management on top of Celery."""

from .cache import Cache
from .codename import selinon_version_codename
from .config import Config
from .data_storage import DataStorage, SelinonMissingDataException
from .dispatcher import Dispatcher
from .errors import CacheMissError
from .errors import ConfigNotInitializedError
from .errors import ConfigurationError
from .errors import FatalTaskError
from .errors import FlowError
from .errors import NoParentNodeError
from .errors import RequestError
from .errors import Retry
from .errors import SelectiveNoPathError
from .errors import StorageError
from .errors import UnknownError
from .errors import UnknownFlowError
from .errors import UnknownStorageError
from .run import run_flow
from .run import run_flow_selective
from .selinon_task import SelinonTask
from .storage import Storage
from .storage_pool import StoragePool
from .system_state import SystemState
from .trace import Trace
from .version import selinon_version

__version__ = selinon_version
__version_info__ = __version__.split('.')
__title__ = 'selinon'
__author__ = 'Fridolin Pokorny'
__license__ = 'BSD License 2.0'
__copyright__ = 'Copyright 2018 Fridolin Pokorny'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Selinon adapter for Redis database."""

import os
import json
from selinon.data_storage import SelinonMissingDataException

from selinon import DataStorage

try:
    import redis
except ImportError as exc:
    raise ImportError("Please install dependencies using `pip3 install selinon[redis]` "
                      "in order to use RedisStorage") from exc


class Redis(DataStorage):  # pylint: disable=too-many-instance-attributes
    """Selinon adapter for Redis database."""

    def __init__(self, host=None, port=6379, db=0, password=None, socket_timeout=None, connection_pool=None,
                 charset=None, errors=None, unix_socket_path=None):
        """Instantiate Redis database adapter.

        :param host: Redis host
        :param port: Redis port
        :param db: Redis database to be used
        :param password: password to be used
        :param socket_timeout: socket timeout
        :param connection_pool: redis connection pool
        :param charset: connection character set
        :param errors: error treating method
        :param unix_socket_path: path to unix socket, if any
        """
        super().__init__()
        self.conn = None
        self.host = host.format(**os.environ) if host else 'localhost'
        self.port = int(port.format(**os.environ)) if isinstance(port, str) else port
        self.db = int(db.format(**os.environ) if isinstance(db, str) else db)  # pylint: disable=invalid-name
        self.password = password.format(**os.environ) if password else None
        self.socket_timeout = socket_timeout
        self.connection_pool = connection_pool
        self.charset = charset or 'utf-8'
        self.errors = errors or 'strict'
        self.unix_socket_path = unix_socket_path

    def is_connected(self):  # noqa
        return self.conn is not None

    def connect(self):  # noqa
        self.conn = redis.Redis(host=self.host, port=self.port, db=self.db, password=self.password,
                                socket_timeout=self.socket_timeout, connection_pool=self.connection_pool,
                                charset=self.charset, errors=self.errors)

    def disconnect(self):  # noqa
        if self.is_connected():
            self.conn.connection_pool.disconnect()
            self.conn = None

    def retrieve(self, flow_name, task_name, task_id):  # noqa
        assert self.is_connected()  # nosec

        ret = self.conn.get(task_id)

        if ret is None:
            raise FileNotFoundError("Record not found in database")

        record = json.loads(ret.decode(self.charset))

        assert record.get('task_name') == task_name  # nosec
        return record.get('result')

    def store(self, node_args, flow_name, task_name, task_id, result):  # noqa
        assert self.is_connected()  # nosec

        record = {
            'node_args': node_args,
            'flow_name': flow_name,
            'task_name': task_name,
            'task_id': task_id,
            'result': result
        }

        self.conn.set(task_id, json.dumps(record))
        return task_id

    def store_error(self, node_args, flow_name, task_name, task_id, exc_info):  # noqa
        # just to make pylint happy
        raise NotImplementedError()


    def delete(self, flow_name, task_name, task_id):  # noqa
        assert self.is_connected()  # nosec

        ret = self.conn.delete(task_id)

        if ret == 0:
            raise SelinonMissingDataException("Record not found in database")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from selinon.errors import CacheMissError
from selinon.caches import RR
from selinon_test_case import SelinonTestCase


class TestRR(SelinonTestCase):
    """
    Test Random-Replacement Cache
    """
    def test_remove(self):
        cache = RR(max_cache_size=1)

        cache.add("item_id1", "item1", "Task1", "flow1")
        assert cache.get("item_id1", "Task1", "flow1") == "item1"

        cache.add("item_id2", "item2", "Task1", "flow1")
        assert cache.get("item_id2", "Task1", "flow1") == "item2"

        with pytest.raises(CacheMissError):
            cache.get("item_id1", "Task1", "flow1")

    def test_random_remove(self):
        cache = RR(max_cache_size=2)

        cache.add("item_id1", "item1", "Task1", "flow1")
        cache.add("item_id2", "item2", "Task1", "flow1")
        cache.add("item_id3", "item3", "Task1", "flow1")

        with pytest.raises(CacheMissError):
            # at least one should fail
            cache.get("item_id1", "Task1", "flow1")
            cache.get("item_id2", "Task1", "flow1")

        assert cache.get("item_id3", "Task1", "flow1") == "item3"
        assert cache.current_cache_size == 2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import os
import copy
import pytest
from selinon.migrations import Migrator
from selinon.errors import MigrationFlowFail
from selinon.errors import MigrationFlowRetry
from selinon.errors import MigrationSkew
from selinon_test_case import SelinonTestCase


def state_dict(**state_kwargs):
    """A helper to make tests more readable and bullet proof - convert kwargs to a dict."""
    # check required field
    if state_kwargs:
        if 'waiting_edges' not in state_kwargs:
            state_kwargs['waiting_edges'] = []

        if 'triggered_edges' not in state_kwargs:
            state_kwargs['triggered_edges'] = []

        if 'active_nodes' not in state_kwargs:
            state_kwargs['active_nodes'] = []

        if 'finished_nodes' not in state_kwargs:
            state_kwargs['finished_nodes'] = {}

        if 'failed_nodes' not in state_kwargs:
            state_kwargs['failed_nodes'] = {}

        unknown = set(state_kwargs.keys()) ^ {
            'active_nodes',
            'finished_nodes',
            'waiting_edges',
            'failed_nodes',
            'triggered_edges'
        }
        if unknown:
            raise ValueError("Unknown state dictionary configuration provided: %s" % unknown)

    return state_kwargs


def migrate_message_test(flow_name, migration_version, original_state):
    """A migration test (successful) execution. Files are expected to be named based on test name."""
    def decorator(test):
        def wrapper(test_instance):
            migration_dir = test_instance.get_migration_dir(test.__name__)
            migrator = Migrator(migration_dir)
            migrated_state, new_migration_version, tainted = migrator.perform_migration(flow_name,
                                                                                        copy.deepcopy(original_state),
                                                                                        migration_version)
            test(test_instance, original_state, migrated_state, new_migration_version, tainted)
            test_instance.sanity_assert(original_state, migrated_state)

        return wrapper

    return decorator


def migrate_message_exception(exception, flow_name, migration_version, original_state):
    """A migration test (exception) execution. Files are expected to be named based on test name."""
    def decorator(test):
        def wrapper(test_instance):
            migration_dir = test_instance.get_migration_dir(test.__name__)
            migrator = Migrator(migration_dir)

            with pytest.raises(exception) as exc_info:
                migrator.perform_migration(flow_name, original_state, migration_version)

            test(test_instance, exc_info)

        return wrapper

    return decorator


class TestPerformMigration(SelinonTestCase):
    """Test actual migration based on generated migrations."""

    def get_migration_dir(self, test_name):
        """Get configuration files for specific test case - path respects test method naming.

        :param test_name: name of test that is going to be executed
        :return: a path to test migration directory
        """
        # Force structure of test data
        return os.path.join(self.DATA_DIR, 'migrations', 'migration_dirs', test_name[len('test_'):])

    @staticmethod
    def sanity_assert(original_state, migrated_state):
        """Check that state stays untouched except waiting_edges."""
        (original_state or {}).pop('waiting_edges', None)
        (migrated_state or {}).pop('waiting_edges', None)
        assert original_state == migrated_state

    @migrate_message_test('flow1', 0, None)
    def test_zero_migration(self, original_state, migrated_state, new_migration_version, tainted):
        """Test when no migration is defined and message keeps migration of version 0."""
        assert original_state == migrated_state
        assert new_migration_version == 0
        assert tainted is False

    @migrate_message_test('flow1', None, None)
    def test_no_migration_version(self, original_state, migrated_state, new_migration_version, tainted):
        """Test no migration version in message - this can occur in case of new flow is scheduled."""
        assert original_state == migrated_state
        assert new_migration_version == 0
        assert tainted is False

    @migrate_message_test('flow1', 0, state_dict(waiting_edges=[1], triggered_edges=[0],
                                                 active_nodes=[{"name": "Task1", "id": "id1"}]))
    def test_one_migration(self, original_state, migrated_state, new_migration_version, tainted):
        """Test one single migration from version 0 to version 1."""
        original_state.pop('waiting_edges')
        assert migrated_state.pop('waiting_edges') == [2]
        assert migrated_state == original_state
        assert new_migration_version == 1
        assert tainted is False

    @migrate_message_test('flow1', None, None)
    def test_start_latest_migration(self, original_state, migrated_state, new_migration_version, tainted):
        """Test that if a flow starts and is picked by worker with some migration, this migration has to be tracked."""
        assert original_state == migrated_state
        assert new_migration_version == 1
        assert tainted is False

    @migrate_message_test('flow1', 1, state_dict(active_nodes=[{'name': 'Task2', 'id': 'id2'}],
                                                 waiting_edges=[1],
                                                 finished_nodes={'Task1': 'id1'}))
    def test_migration_chaining(self, original_state, migrated_state, new_migration_version, tainted):
        assert 'waiting_edges' in migrated_state
        # Edge 2 is added as it waits for Task2 to finish
        assert set(migrated_state.pop('waiting_edges')) == {1, 2}
        original_state.pop('waiting_edges')
        # The edge to Task3 is discarded but as Task2 is in progress, the edge to Task4 is added.
        assert migrated_state == original_state
        assert new_migration_version == 3
        assert tainted is False

    @migrate_message_test('flow1', 1, state_dict(active_nodes=[{'name': 'Task1', 'id': 'id1'}],
                                                 waiting_edges=[1]))
    def test_migration_chaining_no_change(self, original_state, migrated_state, new_migration_version, tainted):
        assert migrated_state == original_state
        assert new_migration_version == 3
        assert tainted is False

    @migrate_message_test('flow1', 0, state_dict(active_nodes=[{'name': 'Task3', 'id': 'id3'}],
                                                 waiting_edges=[1, 2],
                                                 finished_nodes={'Task1': ['id1'], 'Task2': ['id2']}))
    def test_migration_chaining_tainted(self, original_state, migrated_state, new_migration_version, tainted):
        # The second edge is removed in the first migration and later a new one to Task4 is introduced.
        assert migrated_state.pop('waiting_edges') == [1]
        original_state.pop('waiting_edges')
        assert migrated_state == original_state
        assert new_migration_version == 2
        assert tainted is True

    @migrate_message_exception(MigrationSkew, 'flow1', 0, state_dict(active_nodes=[{'name': 'Task1', 'id': 'id1'}]))
    def test_migration_skew(self, _):
        """Test signalizing migration skew - migration version is from future i.e. version file not present."""

    @migrate_message_exception(MigrationFlowRetry, 'flow1', 1,
                               state_dict(active_nodes=[{'name': 'Task3', 'id': 'id3'}],
                                          waiting_edges=[1, 2],
                                          finished_nodes={'Task1': ['id1'], 'Task2': ['id2']}))
    def test_migration_flow_retry(self, exc_info):
        """Check that the first migration causing failure does not allow to continue (tainting nodes)."""
        assert exc_info.value.migration_version == 2
        assert exc_info.value.latest_migration_version == 3
        assert exc_info.value.tainting_nodes == ["Task2"]
        assert exc_info.value.tainted_edge is None

    @migrate_message_exception(MigrationFlowRetry, 'flow1', 1,
                               state_dict(active_nodes=[{'name': 'Task3', 'id': 'id3'}],
                                          waiting_edges=[1, 2],
                                          finished_nodes={'Task1': ['id1'], 'Task2': ['id2']}))
    def test_migration_flow_retry_chained(self, exc_info):
        """The second migration in the chain should raise a failure (tainted edges)."""
        assert exc_info.value.migration_version == 2
        assert exc_info.value.latest_migration_version == 3
        assert exc_info.value.tainting_nodes == ['Task2']
        assert exc_info.value.tainted_edge is None

    @migrate_message_exception(MigrationFlowFail, 'flow1', 1,
                               state_dict(active_nodes=[{'name': 'Task3', 'id': 'id3'}],
                                          waiting_edges=[1, 2],
                                          finished_nodes={'Task1': ['id1'], 'Task2': ['id2']}))
    def test_migration_flow_fail(self, exc_info):
        """Check that the first migration causing failure does not allow to continue (tainting nodes)."""
        assert exc_info.value.migration_version == 2
        assert exc_info.value.latest_migration_version == 3
        assert exc_info.value.tainting_nodes == ["Task2"]
        assert exc_info.value.tainted_edge is None

    @migrate_message_exception(MigrationFlowFail, 'flow1', 1,
                               state_dict(active_nodes=[{'name': 'Task3', 'id': 'id3'}],
                                          waiting_edges=[1, 2],
                                          finished_nodes={'Task1': ['id1'], 'Task2': ['id2']}))
    def test_migration_flow_fail_chained(self, exc_info):
        """The second migration in the chain should raise a failure (tainted edges)."""
        assert exc_info.value.migration_version == 2
        assert exc_info.value.latest_migration_version == 3
        assert exc_info.value.tainting_nodes == ['Task2']
        assert exc_info.value.tainted_edge is None



#!/usr/bin/env python3
# pragma: no cover

from functools import reduce
from urllib.parse import urlparse


def fieldUrlPath(message, key, path):
    try:
        val = reduce(lambda m, k: m[k], key if isinstance(key, list) else [key], message)
        return urlparse(val).path == path
    except:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""A simple filesystem storage implementation."""

import json
import os

from selinon import DataStorage, SelinonMissingDataException


class Filesystem(DataStorage):
    """Selinon adapter for storing task results in a directory."""

    def __init__(self, path=None):
        """Instantiate Filesystem adapter.

        :param path: path to directory to be used
        :type path: str
        """
        super().__init__()
        self.path = (path or '{PWD}').format(**os.environ)
        self._connected = False

    def _construct_base_path(self, flow_name, task_name):
        return os.path.join(self.path, flow_name, task_name)

    def _construct_path(self, flow_name, task_name, task_id):
        return os.path.join(self._construct_base_path(flow_name, task_name), '{}.json'.format(task_id))

    def is_connected(self):
        return self._connected

    def connect(self):
        if not os.path.isdir(self.path):
            os.makedirs(self.path, exist_ok=True)
        self._connected = True

    def disconnect(self):
        pass

    def retrieve(self, flow_name, task_name, task_id):
        path = self._construct_path(flow_name, task_name, task_id)
        with open(path, 'r') as result_file:
            return json.load(result_file)

    def store(self, node_args, flow_name, task_name, task_id, result):  # noqa
        base_path = self._construct_base_path(flow_name, task_name)
        if not os.path.isdir(base_path):
            os.makedirs(base_path, exist_ok=True)

        path = self._construct_path(flow_name, task_name, task_id)
        with open(path, 'w') as result_file:
            json.dump(result, result_file)
        return path

    def store_error(self, node_args, flow_name, task_name, task_id, exc_info):  # noqa
        # just to make pylint happy
        raise NotImplementedError()

    def delete(self, flow_name, task_name, task_id):
        path = self._construct_path(flow_name, task_name, task_id)
        if os.path.exists(path):
            os.remove(path)
        else:
            raise SelinonMissingDataException



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import os
import pytest
from selinon import Config
from selinon import Trace
from selinon_test_case import SelinonTestCase


class TestTrace(SelinonTestCase):
    def test_trace_setup(self):
        assert len(Trace._trace_functions) == 0

        test_file = os.path.join(self.DATA_DIR, 'test_trace_setup.yaml')
        Config.set_config_yaml(test_file, flow_definition_files=[test_file])

        # logging, sentry, storage and a custom function - 3 callbacks registered
        # By default there is configured logging module to be used with Sentry.
        assert len(Trace._trace_functions) == 3
        assert Trace._logger is not None

    @pytest.mark.skip(reason="trace calls are currently not tested")
    def test_trace_call(self):
        # TODO: add tests for actual trace call
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

from selinon_test_case import SelinonTestCase
from selinon import SystemState, Dispatcher, Config


class TestPropagate(SelinonTestCase):
    def test_propagate_node_args_true(self):
        #
        # flow1:
        #
        #     flow2
        #
        # Note:
        #    Arguments are propagated to flow2
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2'], 'condition': self.cond_true}],
            'flow2': []
        }
        self.init(edge_table, propagate_node_args={'flow1': True})

        node_args = {'foo': 'bar'}
        system_state = SystemState(id(self), 'flow1', node_args=node_args)
        retry = system_state.update()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'flow2' in self.instantiated_flows

        flow2 = self.get_flow('flow2')
        assert flow2.node_args == node_args

    def test_propagate_node_args_flow(self):
        #
        # flow1:
        #
        #     flow2  flow3
        #
        # Note:
        #    Arguments are propagated to flow2 but not to flow3
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2', 'flow3'], 'condition': self.cond_true}],
            'flow2': [],
            'flow3': []
        }
        self.init(edge_table, propagate_node_args={'flow1': ['flow2']})

        node_args = {'foo': 'bar'}
        system_state = SystemState(id(self), 'flow1', node_args=node_args)
        retry = system_state.update()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'flow2' in self.instantiated_flows
        assert 'flow3' in self.instantiated_flows

        flow2 = self.get_flow('flow2')
        assert flow2.node_args == node_args

        flow3 = self.get_flow('flow3')
        assert flow3.node_args is None

    def test_propagate_parent_true(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     flow2
        #
        # Note:
        #    Parent are propagated to flow2
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true}],
            'flow2': []
        }
        self.init(edge_table, propagate_parent={'flow1': True})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'flow2' not in self.instantiated_flows
        assert 'Task1' in self.instantiated_tasks

        task1 = self.get_task('Task1')
        self.set_finished(task1, 0xDEADBEEF)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        assert retry is not None
        assert system_state.node_args is None
        assert 'flow2' in self.instantiated_flows

        flow2 = self.get_flow('flow2')

        assert flow2.node_args is None
        assert 'Task1' in flow2.parent
        assert flow2.parent['Task1'] == task1.task_id

    def test_propagate_parent_flow(self):
        #
        # flow1:
        #         Task1
        #           |
        #        -------
        #       |       |
        #     flow2   flow3
        #
        # Note:
        #    Arguments are propagated to flow2 but not to flow3. Result of Task1 is not propagated to flow2 nor flow3.
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2', 'flow3'], 'condition': self.cond_true}],
            'flow2': [],
            'flow3': []
        }
        self.init(edge_table, propagate_parent={'flow1': ['flow2']}, propagate_node_args={'flow1': ['flow2']})

        args = 123
        system_state = SystemState(id(self), 'flow1', args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' not in self.instantiated_flows
        assert 'flow3' not in self.instantiated_flows

        task1 = self.get_task('Task1')
        self.set_finished(task1, 0xDEADBEEF)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        assert retry is not None
        assert 'flow2' in self.instantiated_flows
        assert 'flow3' in self.instantiated_flows

        flow2 = self.get_flow('flow2')
        assert flow2.node_args == args
        assert 'Task1' in flow2.parent
        assert flow2.parent['Task1'] == task1.task_id

        flow3 = self.get_flow('flow3')
        assert flow3.node_args is None
        assert flow3.parent is None

    def test_propagate_parent(self):
        #
        # flow1:
        #
        #     Task1       Task2
        #       |           |
        #     flow2         |
        #       |           |
        #        -----------
        #             |
        #           TaskX
        #
        # flow2:
        #    Run explicitly, but result finished_nodes is:
        #         {'flow3': [<flow3-id>], 'Task2': [<task2-id1>], 'Task3': [<task3-id1>, <task3-id2>]}
        # flow3
        #    Run explicitly, but result finished_nodes is:
        #         {'Task2': [<task2-id2>], 'Task4': [<task4-id1>, <task4-id2>]}
        #
        # We are propagating finished, so we should inspect parent
        #
        edge_table = {
            'flow1': [{'from': ['flow2', 'Task2'], 'to': ['TaskX'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}],
            'flow2': [],
            'flow3': []
        }
        self.init(edge_table, propagate_parent={'flow1': True}, propagate_finished={'flow1': True})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict_1 = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'flow2' not in self.instantiated_flows

        # Task1 and Task2 have finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, None)

        task2 = self.get_task('Task2')
        self.set_finished(task2, None)

        system_state = SystemState(id(self), 'flow1', state=state_dict_1, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict_1 = system_state.to_dict()

        assert retry is not None
        assert 'flow2' in self.instantiated_flows

        # Create flow3 manually
        flow3 = Dispatcher().apply_async(kwargs={'flow_name': 'flow3'}, queue=Config.dispatcher_queues['flow3'])
        self.get_task_instance.register_node(flow3)

        flow2 = self.get_flow('flow2')
        self.set_finished(flow2)
        flow2_result = {'finished_nodes': {'flow3': [flow3.task_id],
                                           'Task2': ['<task2-id1>'],
                                           'Task3': ['<task3-id1>', '<task3-id2>']},
                        'failed_nodes': {}
                        }
        self.set_finished(flow2, flow2_result)

        flow3_result = {'finished_nodes':
                            {'Task2': ['<task2-id2>'], 'Task4': ['<task4-id1>', '<task4-id2>']},
                        'failed_nodes': {}
                        }
        self.set_finished(flow3, flow3_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict_1, node_args=system_state.node_args)
        retry = system_state.update()

        assert retry is not None

        task_x = self.get_task('TaskX')

        task_x_parent = {'Task2': task2.task_id,
                         'flow2': {'Task2': ['<task2-id1>'],
                                   'Task3': ['<task3-id1>', '<task3-id2>'],
                                   'flow3': {'Task2': ['<task2-id2>'],
                                             'Task4': ['<task4-id1>', '<task4-id2>'],
                                             }
                                   }
                         }
        assert task_x.parent == task_x_parent

    def test_propagate_parent_2(self):
        #
        # flow1:
        #
        #     Task1       Task2
        #       |           |
        #     flow2         |
        #       |           |
        #        -----------
        #             |
        #           TaskX
        #
        # flow2:
        #    Not run explicitly, but result finished_nodes is:
        #         {'flow3': [<flow3-id>], 'Task2': [<task2-id1>], 'Task3': [<task3-id1>, <task3-id2>]}
        # flow3:
        #    Not run explicitly, but result finished_nodes is:
        #         {'Task2': [<task2-id2>], 'Task4': [<task4-id1>, <task4-id2>]}
        #
        # We are propagating finished, so we should inspect parent
        #
        edge_table = {
            'flow1': [{'from': ['flow2', 'Task2'], 'to': ['TaskX'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}],
            'flow2': [],
            'flow3': []
        }
        # Make sure propagate_finished is set to False as they are disjoint with propagate_compound_finished;
        # this is checked in selinon
        self.init(edge_table, propagate_parent=dict.fromkeys(edge_table.keys(), True),
                  propagate_finished={'flow1': False},
                  propagate_compound_finished={'flow1': True})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict_1 = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'flow2' not in self.instantiated_flows

        # Task1 and Task2 have finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, None)

        task2 = self.get_task('Task2')
        self.set_finished(task2, None)

        system_state = SystemState(id(self), 'flow1', state=state_dict_1, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict_1 = system_state.to_dict()

        assert retry is not None
        assert 'flow2' in self.instantiated_flows

        # Create flow3 manually
        flow3 = Dispatcher().apply_async(kwargs={'flow_name': 'flow3'}, queue=Config.dispatcher_queues['flow3'])
        self.get_task_instance.register_node(flow3)

        flow2 = self.get_flow('flow2')
        self.set_finished(flow2)
        flow2_result = {'finished_nodes': {'flow3': [flow3.task_id],
                                           'Task2': ['<task2-id1>'],
                                           'Task3': ['<task3-id1>', '<task3-id2>']},
                        'failed_nodes': {}
                        }
        self.set_finished(flow2, flow2_result)

        flow3_result = {'finished_nodes': {'Task2': ['<task2-id2>'], 'Task4': ['<task4-id1>', '<task4-id2>']},
                        'failed_nodes': {}
                        }
        self.set_finished(flow3, flow3_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict_1, node_args=system_state.node_args)
        retry = system_state.update()

        assert retry is not None

        task_x = self.get_task('TaskX')

        task_x_parent = {'Task2': task2.task_id,
                         'flow2': {'Task2': ['<task2-id1>', '<task2-id2>'],
                                   'Task3': ['<task3-id1>', '<task3-id2>'],
                                   'Task4': ['<task4-id1>', '<task4-id2>']
                                   }
                         }

        # we have to check this as a set due to dict randomization
        assert 'Task2' in task_x.parent
        assert task_x.parent['Task2'] == task_x_parent['Task2']

        assert 'flow2' in task_x.parent
        assert 'Task2' in task_x.parent['flow2']
        assert 'Task3' in task_x.parent['flow2']
        assert 'Task4' in task_x.parent['flow2']

        assert set(task_x.parent['flow2']['Task2']) == set(task_x_parent['flow2']['Task2'])
        assert set(task_x.parent['flow2']['Task3']) == set(task_x_parent['flow2']['Task3'])
        assert set(task_x.parent['flow2']['Task4']) == set(task_x_parent['flow2']['Task4'])

    def test_propagate_compound_mixed(self):
        #
        # flow1:
        #
        #     flow2       flow3
        #       |           |
        #        -----------
        #             |
        #           TaskX
        #
        # flow2:
        #    Not run explicitly, but result finished_nodes is:
        #         {'Task2': [<task2-id21>],
        #          'Task3': [<task3-id21>, <task3-id22>]}
        #          'flow4': {'Task2': [<task2-id41>]}
        # flow3:
        #    Not run explicitly, but result finished_nodes is:
        #         {'Task2': [<task2-id32>],
        #          'Task3': [<task3-id31>, <task3-id32>],
        #          'flow4': {'Task2': [<task2-id42>]}
        #
        # We are propagating finished, so we should inspect parent
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2', 'flow3'], 'condition': self.cond_true},
                      {'from': ['flow2', 'flow3'], 'to': ['TaskX'], 'condition': self.cond_true}],
            'flow2': [],
            'flow3': [],
            'flow4': []
        }
        # Make sure propagate_finished is negated propagate_compound_finished
        # this is checked in selinon
        self.init(edge_table,
                  propagate_finished={'flow1': ['flow2']},
                  propagate_compound_finished={'flow1': ['flow3']})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert 'flow2' in self.instantiated_flows

        # Create flow4 manually, we will reuse it, but we pretend that there are 2 instances - one run in flow2
        # another one in flow3
        flow4 = Dispatcher().apply_async(kwargs={'flow_name': 'flow4'}, queue=Config.dispatcher_queues['flow4'])
        self.get_task_instance.register_node(flow4)
        self.set_finished(flow4, {'finished_nodes': {'Task2': ['<task2-id41']}, 'failed_nodes': {}})

        # Create results of flow2 and flow3 manually but they are instantiated by dispatcher
        flow2_result = {'finished_nodes': {'Task2': ['<task2-id21>'],
                                           'Task3': ['<task3-id21>', '<task3-id22>'],
                                           'flow4': [flow4.task_id]},
                        'failed_nodes': {}
                        }
        flow2 = self.get_flow('flow2')
        self.set_finished(flow2, flow2_result)

        flow3_result = {'finished_nodes': {'Task2': ['<task2-id32>'],
                                           'Task3': ['<task3-id31>', '<task3-id32>'],
                                           'flow4': [flow4.task_id]},
                        'failed_nodes': {}
                        }
        flow3 = self.get_flow('flow3')
        self.set_finished(flow3, flow3_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        assert retry is not None
        assert 'TaskX' in self.instantiated_tasks

        task_x_parent = {'flow2': {'Task2': ['<task2-id21>'],
                                   'Task3': ['<task3-id21>', '<task3-id22>'],
                                   'flow4': {'Task2': ['<task2-id41']}},
                         # flow4 is hidden in flow3 as it is compound, see 'Task2'
                         'flow3': {'Task2': ['<task2-id41', '<task2-id32>'],
                                   'Task3': ['<task3-id31>', '<task3-id32>']}}
        task_x = self.get_task('TaskX')

        # we have to check this as a set due to dict randomization
        assert 'flow2' in task_x.parent
        assert 'flow3' in task_x.parent

        assert 'Task2' in task_x.parent['flow2']
        assert 'Task3' in task_x.parent['flow2']
        assert set(task_x.parent['flow2']['Task2']) == set(task_x_parent['flow2']['Task2'])
        assert set(task_x.parent['flow2']['Task3']) == set(task_x_parent['flow2']['Task3'])

        assert 'Task2' in task_x.parent['flow3']
        assert 'Task3' in task_x.parent['flow3']
        assert set(task_x.parent['flow3']['Task2']) == set(task_x_parent['flow3']['Task2'])
        assert set(task_x.parent['flow3']['Task3']) == set(task_x_parent['flow3']['Task3'])

        assert 'flow4' in task_x.parent['flow2']
        assert set(task_x.parent['flow2']['flow4']) == set(task_x_parent['flow2']['flow4'])





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################


class StorageTaskNameMock:
    """Mock storage_task_name dict from configuration"""
    def __init__(self):
        pass

    def __getitem__(self, item):
        # Return what we have asked - try to pretend {'TaskX': 'TaskX'} for each task
        return item



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Task and flow failure handling."""

from itertools import chain

from .errors import ConfigurationError
from .failure_node import FailureNode
from .helpers import check_conf_keys


class Failures:
    """Node failures and fallback handling."""

    def __init__(self, raw_definition, system, flow, last_allocated=None, starting_nodes=None, predicates=None):
        """Construct failures based on definition stated in YAML config files.

        :param raw_definition: raw definition of failures
        :param system: system context
        :param last_allocated: last allocated starting node for linked list
        :param starting_nodes: starting nodes for failures
        :param predicates: all predicates that are used
        """
        self.waiting_nodes = []
        self.fallback_nodes = []

        for failure in raw_definition:
            waiting_nodes_entry = []
            for node_name in failure['nodes']:
                node = system.node_by_name(node_name, graceful=True)
                if not node:
                    raise ConfigurationError("No such node with name '%s' in failure, flow '%s'"
                                             % (node_name, flow.name))
                waiting_nodes_entry.append(node)

            if isinstance(failure['fallback'], list):
                fallback_nodes_entry = []
                for node_name in failure['fallback']:
                    node = system.node_by_name(node_name, graceful=True)
                    if not node:
                        raise ConfigurationError("No such node with name '%s' in failure fallback, flow '%s'"
                                                 % (node_name, flow.name))
                    fallback_nodes_entry.append(node)
            elif isinstance(failure['fallback'], bool):
                fallback_nodes_entry = failure['fallback']
            else:
                raise ConfigurationError("Unknown fallback definition in flow '%s', failure: %s"
                                         % (flow.name, failure))

            self.waiting_nodes.append(waiting_nodes_entry)
            self.fallback_nodes.append(fallback_nodes_entry)

        self.raw_definition = raw_definition
        self.last_allocated = last_allocated
        self.starting_nodes = starting_nodes
        self.predicates = predicates
        self.flow = flow

    def all_waiting_nodes(self):
        """Compute all nodes that have defined a fallback.

        :return: all nodes that for which there is defined a callback
        """
        return list(set(chain(*self.waiting_nodes)))

    def all_fallback_nodes(self):
        """Compute all fallback nodes.

        :return: all nodes that are used as fallback nodes
        """
        # remove True/False flags
        nodes = []
        if isinstance(self.fallback_nodes, bool):
            return nodes

        for fallback in self.fallback_nodes:
            if isinstance(fallback, bool):
                continue

            for node in fallback:
                if not isinstance(node, bool):
                    nodes.append(node)

        return list(set(nodes))

    @staticmethod
    def construct(system, flow, failures_dict):
        """Construct Failres.

        :param system: system context
        :param flow: a flow to which failures conform
        :param failures_dict: construct failures from failures dict
        :rtype: Failures
        """
        for failure in failures_dict:
            if 'nodes' not in failure or failure['nodes'] is None:
                raise ConfigurationError("Failure should state nodes for state 'nodes' to fallback from in flow '%s'"
                                         % flow.name)

            if 'fallback' not in failure:
                raise ConfigurationError("No fallback stated in failure in flow '%s'" % flow.name)

            if not isinstance(failure['nodes'], list):
                failure['nodes'] = [failure['nodes']]

            if not isinstance(failure['fallback'], list) and failure['fallback'] is not True:
                failure['fallback'] = [failure['fallback']]

            if failure['fallback'] is not True and len(failure['fallback']) == 1 and len(failure['nodes']) == 1 \
                    and failure['fallback'][0] == failure['nodes'][0]:
                raise ConfigurationError("Detect cyclic fallback dependency in flow %s, failure on %s"
                                         % (flow.name, failure['nodes'][0]))

            known_conf_opts = ('nodes', 'fallback', 'propagate_failure', 'condition')
            unknown_conf = check_conf_keys(failure, known_conf_opts=known_conf_opts)
            if unknown_conf:
                raise ConfigurationError("Unknown configuration option supplied in fallback definition: %s"
                                         % unknown_conf)

        last_allocated, starting_nodes, predicates = FailureNode.construct(flow, system, failures_dict)
        return Failures(failures_dict, system, flow, last_allocated, starting_nodes, predicates)

    @staticmethod
    def starting_nodes_name(flow_name):
        """Create a starting node name for graph of all failures nodes for generated Python config.

        :param flow_name: flow name for which the starting node should be created
        :return: variable name
        :rtype: str
        """
        return "_%s_failure_starting_nodes" % flow_name

    @staticmethod
    def failure_node_name(flow_name, failure_node):
        """Create a failure node name representation for generated Python config.

        :param flow_name: name of flow for which the representation should be created
        :param failure_node: a node from graph of all failure permutations
        :return: variable name
        :rtype: str
        """
        return "_%s_fail_%s" % (flow_name, "_".join(failure_node.traversed))

    def fallback_nodes_names(self):
        """Compute names for all nodes that are started by fallbacks.

        :return: names of nodes that are started by fallbacks in all failures
        """
        ret = []

        failure_node = self.last_allocated
        while failure_node:
            if isinstance(failure_node.fallback, list):
                ret.extend(failure_node.fallback)
            failure_node = failure_node.failure_link

        return ret

    def waiting_nodes_names(self):
        """Compute all nodes that have defined fallbacks.

        :return: names of all nodes that we are expecting to fail for fallbacks
        """
        return list(self.starting_nodes.keys())

    def dump_all_conditions2stream(self, stream):
        """Dump all condition sources that are present to a stream.

        :param stream: output stream to dump to
        """
        fail_node = self.last_allocated

        while fail_node:
            for idx, predicate in enumerate(fail_node.predicates):
                condition_name = FailureNode.construct_condition_name(fail_node, idx)
                condition_source = predicate.to_source()

                stream.write('def {}(db, node_args):\n'.format(condition_name))
                stream.write('    return {}\n\n\n'.format(condition_source))

            fail_node = fail_node.failure_link

    def dump2stream(self, stream):
        """Dump failures to the Python config file for Dispatcher.

        :param stream: output stream to dump to
        """
        fail_node = self.last_allocated

        while fail_node:
            next_dict = {}
            for key, value in fail_node.next.items():
                next_dict[key] = self.failure_node_name(self.flow.name, value)

            stream.write("%s = {'next': " % self.failure_node_name(self.flow.name, fail_node))

            # print "next_dict"
            stream.write('{')
            printed = False
            for key, value in next_dict.items():
                if printed:
                    stream.write(", ")
                stream.write("'%s': %s" % (key, value))
                printed = True
            stream.write('}, ')

            conditions = []
            condition_strs = []
            for idx, predicate in enumerate(fail_node.predicates):
                conditions.append(FailureNode.construct_condition_name(fail_node, idx))
                condition_strs.append(str(predicate).replace('\'', '\\\''))

            # now list of nodes that should be started in case of failure (fallback)
            stream.write("'fallback': %s, 'propagate_failure': %s, 'conditions': %s, 'condition_strs': %s}\n"
                         % (fail_node.fallbacks, fail_node.propagate_failures,
                            str(conditions).replace("'", ""), condition_strs))
            fail_node = fail_node.failure_link

        stream.write("\n%s = {" % self.starting_nodes_name(self.flow.name))

        printed = False
        for key, value in self.starting_nodes.items():
            if printed:
                stream.write(",")
            stream.write("\n    '%s': %s" % (key, self.failure_node_name(self.flow.name, value)))
            printed = True
        stream.write("\n}\n\n")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Edge representation in task/flow dependency graph."""

from .errors import ConfigurationError
from .helpers import check_conf_keys
from .predicate import Predicate


class Edge:
    """Edge representation."""

    _DEFAULT_RUN_SUBSEQUENT = False
    _DEFAULT_FOLLOW_SUBFLOWS = False

    def __init__(self, nodes_from, nodes_to, predicate, flow, foreach, selective):
        """Initialize edge definition.

        :param nodes_from: nodes from where edge starts
        :type nodes_from: List[Node]
        :param nodes_to: nodes where edge ends
        :type nodes_to: List[Node]
        :param predicate: predicate condition
        :type predicate: Predicate
        :param flow: flow to which edge belongs to
        :type flow: Flow
        :param foreach: foreach defining function and import over which we want to iterate
        :type foreach: dict
        :param selective: selective run flow configuration
        :type selective: None|dict
        """
        self.nodes_from = nodes_from
        self.nodes_to = nodes_to
        self.predicate = predicate
        self.flow = flow
        self.foreach = foreach
        self.selective = selective

    def check(self):
        """Check edge consistency."""
        if self.foreach and self.foreach['propagate_result']:
            # We can propagate result of our foreach function only if:
            #  1. all nodes to are flows
            #  2. propagate_node_args is not set for flows listed in nodes to
            for node_to in self.nodes_to:
                if not node_to.is_flow():
                    raise ConfigurationError("Flag propagate_result listed in foreach configuration in flow '%s' "
                                             "requires all nodes to to be flows, but '%s' is a task"
                                             % (self.flow.name, node_to.name))

                if (isinstance(self.flow.propagate_node_args, bool) and self.flow.propagate_node_args) \
                        or \
                    (isinstance(self.flow.propagate_node_args, list) and node_to.name in self.flow.propagate_node_args):
                    raise ConfigurationError("Cannot propagate node arguments to subflow when propagate_result "
                                             "is set in foreach definition in flow '%s' for node to '%s'"
                                             % (self.flow.name, node_to.name))

        for node in self.predicate.nodes_used():
            if not node.is_task():
                continue

            if node.storage_readonly and self.predicate.requires_message():
                raise ConfigurationError("Cannot inspect results of node '%s' in flow '%s' as this node is "
                                         "configured with readonly storage, condition: %s"
                                         % (node.name, self.flow.name, str(self.predicate)))

        if self.selective:
            # TODO: there are no checks whether tasks that we want to run are actually run - this will be runtime error
            if len(self.nodes_to) > 1:
                raise ConfigurationError("Cannot run selective flows on edges with more than 1 destination nodes in"
                                         "flow '%s'" % self.flow.name)

            if not self.nodes_to[0].is_flow():
                raise ConfigurationError("Cannot run selective flow in flow '%s', destination node '%s' is not a flow,"
                                         % (self.flow.name, self.nodes_to[0].name))

    def foreach_str(self):
        """Construct string representation for foreach node.

        :return: text representation of foreach
        """
        if self.foreach:
            return "foreach %s.%s" % (self.foreach['import'], self.foreach['function'])

        return None

    @classmethod
    def _parse_selective(cls, flow, selective_def):
        """Parse selective edge run definition.

        :param flow: flow in which selective subflow should be run
        :param selective_def: selective flow definition
        :return: parsed selective_flow
        """
        if not selective_def:
            return None

        unknown_conf = check_conf_keys(selective_def, known_conf_opts=('tasks', 'run_subsequent', 'follow_subflows'))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration options supplied for selective flow run in flow '%s': %s"
                                     % (flow.name, unknown_conf))

        if 'tasks' not in selective_def:
            raise ConfigurationError("Configuration for selective subflows expects 'tasks' to be defined in flow '%s'"
                                     % flow.name)

        if not isinstance(selective_def['tasks'], list) or not all(isinstance(t, str) for t in selective_def['tasks']):
            raise ConfigurationError("Configuration for selective subflows expects task names under 'tasks' key in "
                                     "flow '%s', got '%s' instead" % (flow.name, selective_def['tasks']))

        if 'run_subsequent' not in selective_def:
            selective_def['run_subsequent'] = cls._DEFAULT_RUN_SUBSEQUENT

        if 'follow_subflows' not in selective_def:
            selective_def['follow_subflows'] = cls._DEFAULT_FOLLOW_SUBFLOWS

        if not isinstance(selective_def['run_subsequent'], bool) and \
                not isinstance(selective_def['run_subsequent'], list):
            raise ConfigurationError("Option 'run_subsequent' requires boolean or list of flow names "
                                     "in which subsequent nodes should be run in flow '%s', got '%s'"
                                     % (flow.name, selective_def['run_subsequent']))

        if not isinstance(selective_def['follow_subflows'], bool):
            raise ConfigurationError("Option 'follow_subflows' expects boolean got '%s' instead in flow '%s'"
                                     % (selective_def['follow_subflows'], flow.name))

        # So we can use directly selective as **kwargs to run_flow_selective
        # TODO: unify with run_flow_selective
        selective_def['task_names'] = selective_def.pop('tasks')

        return selective_def

    @classmethod
    def from_dict(cls, dict_, system, flow):  # pylint: disable=too-many-branches
        """Construct edge from a dict.

        :param dict_: a dictionary from which the system should be created
        :type dict_: dict
        :param system:
        :type system: System
        :param flow: flow to which edge belongs to
        :type flow: Flow
        :return: edge representation
        :rtype: Edge
        """
        if 'from' not in dict_:
            raise ConfigurationError("Edge definition requires 'from' explicitly to be specified, "
                                     "use empty for starting edge")

        # we allow empty list for a starting edge
        if dict_['from']:
            from_names = dict_['from'] if isinstance(dict_['from'], list) else [dict_['from']]
            nodes_from = [system.node_by_name(n) for n in from_names]
        else:
            nodes_from = []

        if 'to' not in dict_ or not dict_['to']:
            raise ConfigurationError("Edge definition requires 'to' specified")

        to_names = dict_['to'] if isinstance(dict_['to'], list) else [dict_['to']]
        nodes_to = [system.node_by_name(n) for n in to_names]

        if 'condition' in dict_:
            predicate = Predicate.construct(dict_.get('condition'), nodes_from, flow)
        else:
            predicate = Predicate.construct_default(flow)

        foreach = None
        if 'foreach' in dict_:
            foreach_def = dict_['foreach']
            if foreach_def is None or 'function' not in foreach_def or 'import' not in foreach_def:
                raise ConfigurationError("Specification of 'foreach' requires 'function' and 'import' to be set "
                                         "in flow '%s', got %s instead" % (flow.name, foreach_def))

            foreach = {
                'function': foreach_def['function'],
                'import': foreach_def['import'],
                'propagate_result': False
            }

            if 'propagate_result' in foreach_def:
                if not isinstance(foreach_def['propagate_result'], bool):
                    raise ConfigurationError("Propagate result should be bool in flow '%s', got %s instead"
                                             % (flow.name, foreach_def['propagate_result']))

                if foreach_def['propagate_result']:
                    foreach['propagate_result'] = True

                # additional checks for 'propagate_result' are done in Edge.check() since we have chicken-egg problem
                # here

            if not isinstance(foreach_def['function'], str):
                raise ConfigurationError("Wrong function name '%s' supplied in foreach section in flow %s"
                                         % (foreach_def['function'], flow.name))

            if not isinstance(foreach_def['import'], str):
                raise ConfigurationError("Wrong import statement '%s' supplied in foreach section in flow %s"
                                         % (foreach_def['import'], flow.name))

        selective = cls._parse_selective(flow, dict_.get('selective'))

        unknown_conf = check_conf_keys(dict_, known_conf_opts=('from', 'to', 'foreach', 'condition', 'selective'))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration options supplied for edge in flow '%s': %s"
                                     % (flow.name, unknown_conf))

        return Edge(
            nodes_from=nodes_from,
            nodes_to=nodes_to,
            predicate=predicate,
            flow=flow,
            foreach=foreach,
            selective=selective
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

from selinon_test_case import SelinonTestCase

from selinon import SystemState


# Let's make it constant, this shouldn't affect tests at all
_FOREACH_COUNT = 20


class TestForeach(SelinonTestCase):
    def test_foreach_start(self):
        #
        # flow1:
        #
        #       |      |             |
        #     Task1  Task1   ...   Task1
        #
        # Note:
        #   There will be spawned _FOREACH_COUNT Task2
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true,
                       'foreach': lambda x, y: range(_FOREACH_COUNT), 'foreach_propagate_result': False}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert len(self.get_all_tasks('Task1')) == _FOREACH_COUNT
        tasks_state_dict = [node for node in state_dict['active_nodes'] if node['name'] == 'Task1']
        assert len(tasks_state_dict) == _FOREACH_COUNT

    def test_foreach_basic(self):
        #
        # flow1:
        #
        #             Task1
        #               |
        #               |
        #       ---------------------
        #       |      |             |
        #       |      |             |
        #     Task2  Task2   ...   Task2
        #
        # Note:
        #   There will be spawned _FOREACH_COUNT Task2
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true,
                       'foreach': lambda x, y: range(_FOREACH_COUNT), 'foreach_propagate_result': False},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, "some result")

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks

        assert len(self.get_all_tasks('Task2')) == _FOREACH_COUNT
        tasks_state_dict = [node for node in state_dict['active_nodes'] if node['name'] == 'Task2']
        assert len(tasks_state_dict) == _FOREACH_COUNT

    def test_foreach_propagate_result(self):
        #
        # flow1:
        #
        #             Task1
        #               |
        #               |
        #       ---------------------
        #       |      |             |
        #       |      |             |
        #     flow2  flow2   ...   flow2
        #
        # Note:
        #   There will be spawned _FOREACH_COUNT flow2, arguments are passed from foreach function
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true,
                       'foreach': lambda x, y: range(_FOREACH_COUNT), 'foreach_propagate_result': True},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
            'flow2': []
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, "some result")

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' in self.instantiated_flows

        tasks_state_dict = [node for node in state_dict['active_nodes'] if node['name'] == 'flow2']
        assert len(tasks_state_dict) == _FOREACH_COUNT

        # Inspect node_args as we set propagate_result for foreach
        all_flow_args = [flow.node_args for flow in self.get_all_flows('flow2')]
        assert all_flow_args == list(range(_FOREACH_COUNT))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from selinon_test_case import SelinonTestCase

from selinon.errors import FlowError
from selinon import SystemState


class TestNodeFailuresCond(SelinonTestCase):
    def test_single_failure_cond_false(self):
        #
        # flow1:
        #
        #     Task1 X .. Task3
        #       |
        #       |
        #     Task2
        #
        # Note:
        #   Task1 will fail, Task2 shouldn't be run
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {'Task1': {'next:': {},
                                'fallback': [['Task2']],
                                'conditions': [self.cond_false],
                                'condition_strs': ['false']}}
        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Task1 has failed
        task1 = self.get_task('Task1')
        self.set_failed(task1, KeyError("Some exception raised"))

        with pytest.raises(FlowError):
            system_state = SystemState(id(self), 'flow1', state=state_dict,
                                       node_args=system_state.node_args)
            system_state.update()

        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks

    def test_single_failure_cond_true(self):
        #
        # flow1:
        #
        #     Task1 X
        #       |
        #       |
        #     Task2
        #
        # Note:
        #   Task1 will fail, Task2 should be run
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {'Task1': {'next:': {},
                                'fallback': [['Task3']],
                                'conditions': [self.cond_true],
                                'condition_strs': ['true']}}
        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Task1 has failed
        task1 = self.get_task('Task1')
        self.set_failed(task1, KeyError("Some exception raised"))

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()

        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert retry is not None

    def test_multi_failures(self):
        #
        # flow1:
        #
        #     Task1 X
        #       |
        #       |
        #     Task2
        #
        # Note:
        #   Task1 will fail, Task2 should be run
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        failures = {
            'flow1': {'Task1': {'next:': {},
                                'fallback': [['Task3'], ['Task4', 'Task5'], ['Task6', 'Task7']],
                                'conditions': [self.cond_true, self.cond_false, self.cond_true],
                                'condition_strs': ['true', 'false', 'true']}}
        }
        self.init(edge_table, failures=failures)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Task1 has failed
        task1 = self.get_task('Task1')
        self.set_failed(task1, KeyError("Some exception raised"))

        system_state = SystemState(id(self), 'flow1', state=state_dict,
                                   node_args=system_state.node_args)
        retry = system_state.update()

        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' not in self.instantiated_tasks
        assert 'Task5' not in self.instantiated_tasks
        assert 'Task6' in self.instantiated_tasks
        assert 'Task7' in self.instantiated_tasks
        assert retry is not None




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""A flow representation."""

import logging

from .cache_config import CacheConfig
from .edge import Edge
from .errors import ConfigurationError
from .failures import Failures
from .helpers import check_conf_keys
from .node import Node
from .strategy import Strategy


class Flow(Node):  # pylint: disable=too-many-instance-attributes
    """Flow representation."""

    _DEFAULT_MAX_RETRY = 0
    _DEFAULT_RETRY_COUNTDOWN = 0
    _logger = logging.getLogger(__name__)

    def __init__(self, name, **opts):
        """Initialize flow node representation.

        :param name: flow name
        :type name: str
        :param opts: additional flow options as provided in YAML configuration, see implementation for more details
        """
        super().__init__(name)

        self._logger.debug("Creating flow '%s'", name)
        self.edges = opts.pop('edges', [])
        self.failures = opts.pop('failures', None)
        self.nowait_nodes = opts.pop('nowait_nodes', [])
        self.node_args_from_first = opts.pop('node_args_from_dict', False)
        self.queue_name = self._expand_queue_name(opts.pop('queue', None))
        self.strategy = Strategy.from_dict(opts.pop('sampling', {}), self.name)

        self.propagate_node_args = opts.pop('propagate_node_args', False)
        self.propagate_parent = opts.pop('propagate_parent', False)
        self.propagate_parent_failures = opts.pop('propagate_parent_failures', False)
        self.propagate_finished = opts.pop('propagate_finished', False)
        self.propagate_compound_finished = opts.pop('propagate_compound_finished', False)
        self.propagate_failures = opts.pop('propagate_failures', False)
        self.propagate_compound_failures = opts.pop('propagate_compound_failures', False)
        self.throttling = self.parse_throttling(opts.pop('throttling', {}))
        self.cache_config = opts.pop('cache_config', CacheConfig.get_default(self.name))
        self.max_retry = opts.pop('max_retry', self._DEFAULT_MAX_RETRY)
        self.retry_countdown = opts.pop('retry_countdown', self._DEFAULT_RETRY_COUNTDOWN)
        self.eager_failures = opts.pop('eager_failures', [])

        # disjoint config options
        assert self.propagate_finished is not True and self.propagate_compound_finished is not True  # nosec
        assert self.propagate_failures is not True and self.propagate_compound_failures is not True  # nosec

        if opts:
            raise ConfigurationError("Unknown flow option provided for flow '%s': %s" % (name, opts))

    def _set_propagate(self, system, flow_def, propagate_type):
        """Parse propagate_node_args flag and adjust flow accordingly.

        :param system: system that is used
        :param flow_def: flow definition
        :param propagate_type: propagate flag type
        """
        ret = False

        if propagate_type in flow_def and flow_def[propagate_type] is not None:
            if not isinstance(flow_def[propagate_type], list) and \
                    not isinstance(flow_def[propagate_type], bool):
                flow_def[propagate_type] = [flow_def[propagate_type]]

            if isinstance(flow_def[propagate_type], list):
                ret = []
                for node_name in flow_def[propagate_type]:
                    node = system.flow_by_name(node_name)
                    ret.append(node)
            elif isinstance(flow_def[propagate_type], bool):
                ret = flow_def[propagate_type]
            else:
                raise ConfigurationError("Unknown value in '%s' in flow %s" % (self.name, propagate_type))

        return ret

    def _check_conf_keys(self, flow_def):
        """Check configuration keys so no unknown and unwanted configuration is supplied.

        :param flow_def: dictionary containing flow definition
        :raises: ValueError on wrong configuration
        """
        known_conf_keys = ('name', 'failures', 'nowait', 'cache', 'sampling', 'throttling', 'node_args_from_first',
                           'propagate_node_args', 'propagate_finished', 'propagate_parent', 'propagate_parent_failures',
                           'edges', 'propagate_compound_finished', 'queue', 'max_retry', 'retry_countdown',
                           'propagate_failures', 'propagate_compound_failures', 'eager_failures')

        unknown_conf = check_conf_keys(flow_def, known_conf_keys)
        if unknown_conf:
            raise ConfigurationError("Unknown configuration option for flow '%s' supplied: %s"
                                     % (self.name, unknown_conf))

    def parse_definition(self, flow_def, system):
        """Parse flow definition (fill flow attributes) from a dictionary.

        :param flow_def: dictionary containing flow definition
        :param system: system in which flow is defined
        """
        # pylint: disable=too-many-branches
        assert flow_def['name'] == self.name  # nosec
        self._check_conf_keys(flow_def)

        if self.edges:
            raise ConfigurationError("Multiple definitions of flow '%s'" % self.name)

        if 'edges' not in flow_def.keys() or not flow_def['edges'] or not isinstance(flow_def['edges'], list):
            raise ConfigurationError("No flow edges provided in flow %r" % self.name)

        for edge_def in flow_def['edges']:
            edge = Edge.from_dict(edge_def, system, self)
            self.add_edge(edge)

        if 'failures' in flow_def and flow_def['failures']:
            failures = Failures.construct(system, self, flow_def['failures'])
            self.failures = failures

        if 'nowait' in flow_def and flow_def['nowait'] is not None:
            if not isinstance(flow_def['nowait'], list):
                flow_def['nowait'] = [flow_def['nowait']]

            for node_name in flow_def['nowait']:
                node = system.node_by_name(node_name)
                self.add_nowait_node(node)

        if 'eager_failures' in flow_def and flow_def['eager_failures']:
            if not isinstance(flow_def['eager_failures'], list):
                flow_def['eager_failures'] = [flow_def['eager_failures']]

            for node_name in flow_def['eager_failures']:
                node = system.node_by_name(node_name)
                self.add_eager_failure(node)

        if 'cache' in flow_def:
            if not isinstance(flow_def['cache'], dict):
                raise ConfigurationError("Flow cache for flow '%s' should be a dict with configuration, "
                                         "got '%s' instead" % (self.name, flow_def['cache']))
            self.cache_config = CacheConfig.from_dict(flow_def['cache'], self.name)

        if 'sampling' in flow_def:
            self.strategy = Strategy.from_dict(flow_def.get('sampling'), self.name)

        self.throttling = self.parse_throttling(flow_def.pop('throttling', {}))
        self.node_args_from_first = flow_def.get('node_args_from_first', False)
        self.propagate_node_args = self._set_propagate(system, flow_def, 'propagate_node_args')
        self.propagate_parent = self._set_propagate(system, flow_def, 'propagate_parent')
        self.propagate_parent_failures = self._set_propagate(system, flow_def, 'propagate_parent_failures')
        self.propagate_finished = self._set_propagate(system, flow_def, 'propagate_finished')
        self.propagate_compound_finished = self._set_propagate(system, flow_def, 'propagate_compound_finished')
        self.propagate_failures = self._set_propagate(system, flow_def, 'propagate_failures')
        self.propagate_compound_failures = self._set_propagate(system, flow_def, 'propagate_compound_failures')
        self.queue_name = self._expand_queue_name(flow_def.get('queue'))
        self.max_retry = flow_def.get('max_retry', self._DEFAULT_MAX_RETRY)
        self.retry_countdown = flow_def.get('retry_countdown', self._DEFAULT_RETRY_COUNTDOWN)

    def add_edge(self, edge):
        """Add edge to this flow.

        :param edge: edge to be added
        :type edge: List[Edge]
        """
        self.edges.append(edge)

    def add_nowait_node(self, node):
        """Add edge to this flow, do not wait for node to finish.

        :param node: add a node that should be marked with nowait flag
        """
        self.nowait_nodes.append(node)

    def add_eager_failure(self, node):
        """Add a node to eager failure nodes listing.

        :param node: a node that should be marked as eager failure node
        """
        self.eager_failures.append(node)

    def all_nodes_from(self):
        """Compute all nodes that are stated in 'from' in edges section for this flow.

        :return: all source nodes in flow, excluding failures
        """
        all_nodes_from = set()

        for edge in self.edges:
            all_nodes_from |= set(edge.nodes_from)

        return list(all_nodes_from)

    def all_nodes_to(self):
        """Compute all nodes that are stated in 'to' in edges section for this flow.

        :return: all destination nodes in flow, excluding failures
        """
        all_nodes_to = set()

        for edge in self.edges:
            all_nodes_to |= set(edge.nodes_to)

        return list(all_nodes_to)

    def all_source_nodes(self):
        """Compute all source nodes for this flow.

        :return: all source nodes in flow, including failures
        """
        if self.failures:
            return list(set(self.all_nodes_from()) | set(self.failures.all_waiting_nodes()))

        return self.all_nodes_from()

    def all_destination_nodes(self):
        """Compute all destination nodes for this flow.

        :return: all destination nodes in flow, including failures
        """
        if self.failures:
            return list(set(self.all_nodes_to()) | set(self.failures.all_fallback_nodes()))

        return self.all_nodes_to()

    def all_used_nodes(self):
        """Get all used nodes in this flow, including failures.

        :return: all used nodes in flow
        """
        return list(set(self.all_destination_nodes()) | set(self.all_source_nodes()))

    @staticmethod
    def _should_config(dst_node_name, configuration):
        """Syntax sugar for configuration entries that accept list or boolean at the same time.

        :param dst_node_name: destination node to which configuration should be propagated
        :param configuration: configuration that should be checked
        :return: true if node_name satisfies configuration
        """
        if configuration is True:
            return True

        if isinstance(configuration, list):
            return dst_node_name in configuration

        return False

    def should_propagate_finished(self, dst_node_name):
        """Check whether this flow should propagate info about finished nodes.

        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_finished
        """
        return self._should_config(dst_node_name, self.propagate_finished)

    def should_propagate_failures(self, dst_node_name):
        """Check whether this flow should propagate info about failed nodes.

        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_failures
        """
        return self._should_config(dst_node_name, self.propagate_failures)

    def should_propagate_node_args(self, dst_node_name):
        """Check whether this flow should propagate flow arguments.

        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_node_args
        """
        return self._should_config(dst_node_name, self.propagate_node_args)

    def should_propagate_parent(self, dst_node_name):
        """Check whether this flow should propagate parent nodes.

        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_parent
        """
        return self._should_config(dst_node_name, self.propagate_parent)

    def should_propagate_parent_failures(self, dst_node_name):  # pylint: disable=invalid-name
        """Check whether this flow should propagate parent node failures.

        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_parent_failures
        """
        return self._should_config(dst_node_name, self.propagate_parent_failures)

    def should_propagate_compound_finished(self, dst_node_name):  # pylint: disable=invalid-name
        """Check whether this flow should info about finished nodes (in compound/flattered mode).

        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_compound_finished
        """
        return self._should_config(dst_node_name, self.propagate_compound_finished)

    def should_propagate_compound_failures(self, dst_node_name):  # pylint: disable=invalid-name
        """Check whether this flow should info about failures (in compound/flattered mode).

        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_compound_failures
        """
        return self._should_config(dst_node_name, self.propagate_compound_failures)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import os
import shlex
import pytest

from click.testing import CliRunner

from selinon_test_case import SelinonTestCase

from selinon.codename import selinon_version_codename
from selinon.version import selinon_version
from selinon.cli import execute
from selinon.cli import inspect
from selinon.cli import migrate
from selinon.cli import plot
from selinon.cli import version


class TestCli(SelinonTestCase):
    @staticmethod
    def parse_colon_separated_output(output):
        result = []

        if output[-1] == '\n':
            output = output[:-1]

        for line in output.split('\n'):
            result.append(tuple(line.split(':')))

        return result

    @staticmethod
    def check_result_files(path):
        result_files = []
        for root, _, files in os.walk(path):
            for file in files:
                result_files.append(os.path.join(root, file))

        file_sizes = map(os.path.getsize, result_files)
        assert all(size > 0 for size in file_sizes)

        return result_files

    @pytest.mark.skip
    def test_execute(self):
        # This test is skipped due to failures. To make this work,
        # https://github.com/selinon/selinon/issues/97 needs to be implemented first.
        options = shlex.split("--nodes-definition %s --flow-definitions %s --flow-name flow1 --node-args {} "
                              "--node-args-json --concurrency 1 --sleep-time 0"
                              % (os.path.join(self.DATA_DIR, 'cli', 'execute', 'nodes.yaml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'execute', 'flows')))
        # Filesystem storage expands path to result directory based on $PWD
        runner = CliRunner(env={'PWD': os.getenv('PWD', '.')})

        with runner.isolated_filesystem():
            result = runner.invoke(execute, options)
            assert result.exit_code == 0, result.output

            # Results are stored on filesystem, check that files are actually present and have non-zero size
            result_files = self.check_result_files('flow1')
            assert len(result_files) == 3

    @pytest.mark.skip
    def test_execute_selective_flow(self):
        # This test is skipped due to failures. To make this work,
        # https://github.com/selinon/selinon/issues/97 needs to be implemented first.
        options = shlex.split("--nodes-definition %s --flow-definitions %s --flow-name flow1 --node-args-file %s "
                              "--selective-task-names Task1,Task2 --sleep-time 0"
                              % (os.path.join(self.DATA_DIR, 'cli', 'execute_selective', 'nodes.yaml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'execute_selective', 'flows'),
                                 os.path.join(self.DATA_DIR, 'cli', 'execute_selective', 'node_args.json')))
        # Filesystem storage expands path to result directory based on $PWD
        runner = CliRunner(env={'PWD': os.getenv('PWD', '.')})

        with runner.isolated_filesystem():
            result = runner.invoke(execute, options)
            assert result.exit_code == 0, result.output

            # Results are stored on filesystem, check that files are actually present and have non-zero size
            result_files = self.check_result_files('flow1')
            assert len(result_files) == 2

    def test_inspect(self):
        options = shlex.split("--nodes-definition %s --flow-definitions %s"
                              % (os.path.join(self.DATA_DIR, 'cli', 'inspect', 'nodes.yml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'inspect', 'flows')))
        runner = CliRunner()
        result = runner.invoke(inspect, options)
        assert result.exit_code == 0, result.output

    def test_inspect_list_task_queues(self):
        options = shlex.split("--nodes-definition %s --flow-definitions %s --list-task-queues"
                              % (os.path.join(self.DATA_DIR, 'cli', 'inspect', 'nodes.yml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'inspect', 'flows')))
        runner = CliRunner()
        result = runner.invoke(inspect, options)

        assert result.exit_code == 0, result.output
        output = self.parse_colon_separated_output(result.output)
        assert {
            ('Task1', 'celery'),
            ('Task2', 'celery'),
            ('Task3', 'celery'),
            ('Task4', 'celery')
        } == set(output)

    def test_inspect_list_dispatcher_queues(self):
        options = shlex.split("--nodes-definition %s --flow-definitions %s --list-dispatcher-queues"
                              % (os.path.join(self.DATA_DIR, 'cli', 'inspect', 'nodes.yml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'inspect', 'flows')))
        runner = CliRunner()
        result = runner.invoke(inspect, options)

        assert result.exit_code == 0, result.output
        output = self.parse_colon_separated_output(result.output)
        assert {
            ('flow1', 'flow1_v0'),
            ('flow2', 'celery')
        } == set(output)

    def test_inspect_list_queues(self):
        options = shlex.split("--nodes-definition %s --flow-definitions %s --list-dispatcher-queues --list-task-queues"
                              % (os.path.join(self.DATA_DIR, 'cli', 'inspect', 'nodes.yml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'inspect', 'flows')))
        runner = CliRunner()
        result = runner.invoke(inspect, options)

        assert result.exit_code == 0, result.output
        output = self.parse_colon_separated_output(result.output)
        assert {
            ('flow1', 'flow1_v0'),
            ('flow2', 'celery'),
            ('Task1', 'celery'),
            ('Task2', 'celery'),
            ('Task3', 'celery'),
            ('Task4', 'celery')
        } == set(output)

    def test_migrate(self):
        options = shlex.split("--nodes-definition %s --flow-definitions %s "
                              "--old-nodes-definition %s --old-flow-definitions %s "
                              "--no-meta --migration-dir . --tainted-flows RETRY"
                              % (os.path.join(self.DATA_DIR, 'cli', 'migrate', 'new', 'nodes.yaml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'migrate', 'new', 'flows'),
                                 os.path.join(self.DATA_DIR, 'cli', 'migrate', 'old', 'nodes.yaml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'migrate', 'old', 'flows')))
        runner = CliRunner()

        with runner.isolated_filesystem():
            result = runner.invoke(migrate, options)

            assert result.exit_code == 0, result.output
            output_files = os.listdir('.')
            assert '1.json' in output_files
            assert os.path.getsize('1.json') > 0

    def test_plot(self):
        options = shlex.split("--nodes-definition %s --flow-definitions %s --output-dir . --format png"
                              % (os.path.join(self.DATA_DIR, 'cli', 'plot', 'nodes.yml'),
                                 os.path.join(self.DATA_DIR, 'cli', 'plot', 'flows')))
        runner = CliRunner()

        with runner.isolated_filesystem():
            result = runner.invoke(plot, options)

            assert result.exit_code == 0, result.output
            output_files = os.listdir('.')
            assert {'flow1.png', 'flow2.png'} == set(output_files)

    def test_version(self):
        runner = CliRunner()

        result = runner.invoke(version)
        assert result.exit_code == 0, result.output
        output = self.parse_colon_separated_output(result.output)

        for entry in output:
            assert len(entry) == 2

            if entry[0] == 'Selinon version':
                assert entry[1].strip() == selinon_version
            elif entry[0] == 'Celery version':
                # not sure if Celery is installed, simply pass this
                pass
            else:
                raise ValueError("Unknown version entry in output: %r" % entry[0])

    def test_version_codename(self):
        runner = CliRunner()
        result = runner.invoke(version, ['--codename'])
        assert result.exit_code == 0, result.output
        assert result.output.strip() == selinon_version_codename



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from flexmock import flexmock
from selinon_test_case import SelinonTestCase
from selinon import SystemState
from selinon.errors import DispatcherRetry
from selinon import DataStorage
from celery.result import AsyncResult


class TestFlowRetry(SelinonTestCase):
    class MyStorage(DataStorage):
        def __init__(self):
            pass

        def connect(self):
            # shouldn't be called
            raise NotImplementedError()

        def disconnect(self):
            # called on destruction
            pass

        def is_connected(self):
            return True

        def store(self, node_args, flow_name, task_name, task_id, result):
            raise ValueError("Some exception raised")

        def retrieve(self, flow_name, task_name, task_id):
            raise IndexError("Another exception raised on retrieval")

    def test_retrieve_issue(self):
        #
        # flow1:
        #
        #     Task1
        #
        # Checks that if there is an issue with retrieving task result from task storage in condition, dispatcher
        # retries flow
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': 'Task2', 'condition': lambda db, _: db.get('Task1')}]
        }
        self.init(
            edge_table,
            storage_mapping={'Storage1': self.MyStorage()},
            task2storage_mapping={'Task1': 'Storage1'}
        )

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None

        task1 = self.get_task('Task1')
        self.set_finished(task1, "some result")

        with pytest.raises(DispatcherRetry) as exc_info:
            SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args).update()

        assert exc_info.value.keep_state is True
        assert exc_info.value.adjust_retry_count is False

    def test_result_backend_issue(self):
        #
        # flow1:
        #
        #     Task1
        #
        # Checks that if there is an issue with result backend, dispatcher retries as there is no info on how
        # to continue
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(
            edge_table,
            storage_mapping={'Storage1': self.MyStorage()},
            task2storage_mapping={'Task1': 'Storage1'}
        )
        node_args = {'Foo': 'bar'}

        system_state = SystemState(id(self), 'flow1', node_args=node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None

        task1 = self.get_task('Task1')
        self.set_finished(task1, "some result")

        flexmock(AsyncResult).\
            should_receive('successful').\
            and_raise(ValueError, "Some error raised due to result backed issues")

        with pytest.raises(DispatcherRetry) as exc_info:
            SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args).update()

        assert exc_info.value.keep_state is True
        assert exc_info.value.adjust_retry_count is False




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################


def strategy_function(status):
    """
    Strategy function for system state sampling - simplified version, more robust solutions tested in selinon
    """
    if len(status['active_nodes']) > 0 or len(status['new_started_nodes']) > 0:
        return 2
    else:
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import os
from selinon import Config
from selinon_test_case import SelinonTestCase
from selinon.global_config import GlobalConfig


class TestConfig(SelinonTestCase):
    def test_parse_trace_sentry_without_environ(self):
        expected_dsn = "https://without_environ:secret@sentry.io/project"
        dns = {'dsn': expected_dsn}
        GlobalConfig._parse_trace_sentry(dns)
        assert expected_dsn == GlobalConfig._trace_sentry[0]
        GlobalConfig._trace_sentry.pop()
        
    def test_parse_trace_sentry_with_environ(self):
        expected_dsn = "https://from_environ:secret@sentry.io/project"
        os.environ["DSN"] = expected_dsn
        dns = {'dsn': '{DSN}'}
        GlobalConfig._parse_trace_sentry(dns)
        assert expected_dsn == GlobalConfig._trace_sentry[0]
        GlobalConfig._trace_sentry.pop()


    def test_set_config_yaml_simple(self):
        test_file = os.path.join(self.DATA_DIR, 'test_set_config.yaml')
        Config.set_config_yaml(test_file, flow_definition_files=[test_file])

        # TODO edge table is not tested
        # TODO: make sure we inspect propagated values

        tasks_available = {'task1', 'task2', 'task3'}
        flows_available = {'flow1'}
        storages_available = {'MyStorage'}

        assert tasks_available == set(Config.task_classes)
        assert tasks_available == Config.task_queues.keys()
        assert flows_available == Config.dispatcher_queues.keys()
        assert tasks_available == set(Config.task2storage_mapping.keys())
        assert storages_available == set(Config.storage_mapping.keys())
        assert storages_available == set(Config.storage2storage_cache.keys())
        assert flows_available == set(Config.node_args_from_first.keys())
        assert flows_available == set(Config.propagate_node_args.keys())
        assert flows_available == set(Config.propagate_finished.keys())
        assert flows_available == set(Config.propagate_parent.keys())
        assert flows_available == set(Config.propagate_compound_finished.keys())
        assert tasks_available == set(Config.throttle_tasks.keys())
        assert flows_available == set(Config.throttle_flows.keys())
        assert tasks_available | flows_available == set(Config.max_retry.keys())
        assert tasks_available | flows_available == set(Config.retry_countdown.keys())
        assert tasks_available == set(Config.storage_readonly.keys())
        assert flows_available == set(Config.nowait_nodes.keys())
        assert flows_available == set(Config.strategies.keys())

        assert 'flow1' in Config.failures
        assert {'task1'} == set(Config.nowait_nodes.get('flow1'))
        assert 'schema.json' == Config.output_schemas.get('task1')

    def test_set_config_dict_simple(self):
        nodes = {
            'tasks': [
                {
                    'name': 'Task1',
                    'import': 'testapp.tasks'
                },
                {
                    'name': 'task2',
                    'import': 'testapp.tasks'
                }
            ],
            'flows': [
                'flow1'
            ]
        }

        flows = [
            {
                'flow-definitions': [
                    {
                        'name': 'flow1',
                        'edges': [
                            {
                                'from': None,
                                'to': 'Task1'
                            },
                            {
                                'from': 'Task1',
                                'to': 'task2'
                            }
                        ]
                    }
                ]
            }
        ]

        Config.set_config_dict(nodes, [flows])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from selinon.errors import CacheMissError
from selinon.caches import MRU
from selinon_test_case import SelinonTestCase


class TestMRU(SelinonTestCase):
    """
    Test Most-Recently-Used cache
    """
    @staticmethod
    def _item_id2item(i):
        return "x%s" % i

    def test_one_item_miss(self):
        cache = MRU(max_cache_size=1)

        cache.add("item_id1", "item1", "Task1", "flow1")
        cache.add("item_id2", "item2", "Task1", "flow1")

        with pytest.raises(CacheMissError):
            cache.get("item_id1", "Task1", "flow1")

    def test_two_items(self):
        cache = MRU(max_cache_size=2)

        cache.add("item_id1", "item1", "Task1", "flow1")
        cache.add("item_id2", "item2", "Task1", "flow1")
        cache.add("item_id3", "item3", "Task1", "flow1")

        with pytest.raises(CacheMissError):
            cache.get("item_id2", "Task1", "flow1")

        assert cache.get("item_id1", "Task1", "flow1") == "item1"
        assert cache.get("item_id3", "Task1", "flow1") == "item3"

        # we made "item2" a candidate for next deletion as "item3" was used lastly
        cache.add("item_id1", "item1", "Task1", "flow1")
        with pytest.raises(CacheMissError):
            cache.get("item_id2", "Task1", "flow1")

    def test_multiple_items(self):
        item_count = 16
        cache = MRU(max_cache_size=item_count)

        for item_id in range(item_count):
            cache.add(item_id, self._item_id2item(item_id), "Task1", "flow1")

        cache.add(item_count, self._item_id2item(item_count), "Task1", "flow1")

        with pytest.raises(CacheMissError):
            # the very last is removed
            cache.get(item_count - 1, "Task1", "flow1")

        for item_id in range(item_count - 2, -1, -1):
            assert cache.get(item_id, "Task1", "flow1") == self._item_id2item(item_id)

        assert cache.get(item_count, "Task1", "flow1") == self._item_id2item(item_count)

        # re-add 0
        cache.add(0, self._item_id2item(0), "Task1", "flow1")

        with pytest.raises(CacheMissError):
            cache.get(item_count - 1, "Task1", "flow1")

        for item_id in range(item_count - 1):
            assert cache.get(item_id, "Task1", "flow1") == self._item_id2item(item_id)

        assert cache.get(item_count, "Task1", "flow1") == self._item_id2item(item_count)

    def test_multiple_items_get(self):
        item_count = 16
        cache = MRU(max_cache_size=item_count)

        for item_id in range(item_count):
            cache.add(item_id, self._item_id2item(item_id), "Task1", "flow1")

        # let's use the following 5 items, so they get to the front
        used_items = [10, 5, 3, 1, 7]
        for item_id in used_items:
            assert cache.get(item_id, "Task1", "flow1") == self._item_id2item(item_id)

        # add 5 dummy elements
        for item_id in range(1, 6):
            cache.add(item_id*100, self._item_id2item(item_id*100), "Task1", "flow1")

        with pytest.raises(CacheMissError):
            # last item that was used
            cache.get(used_items[-1], "Task1", "flow1")

        for item_id in range(1, 5):
            with pytest.raises(CacheMissError):
                cache.get(item_id*100, "Task1", "flow1") == self._item_id2item(item_id*100)

        # last inserted when adding 5 dummy elements
        assert cache.get(500, "Task1", "flow1") == self._item_id2item(500)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""A task representation from YAML config file."""

import logging

from .errors import ConfigurationError
from .node import Node
from .selective_run_function import SelectiveRunFunction


class Task(Node):
    # pylint: disable=too-many-instance-attributes,arguments-differ
    """A task representation within the system."""

    _DEFAULT_MAX_RETRY = 0
    _DEFAULT_RETRY_COUNTDOWN = 0
    _logger = logging.getLogger(__name__)

    def __init__(self, name, import_path, storage, **opts):
        """Initialize a task node in a flow graph.

        :param name: name of the task
        :param import_path: tasks's import
        :param storage: storage that should be used
        :param opts: additional options for task
        """
        super().__init__(name)

        self.class_name = opts.pop('classname', name)
        self.storage = storage
        self.import_path = import_path

        if 'storage_task_name' in opts and not self.storage:
            raise ConfigurationError("Unable to assign storage_task_name for task '%s' (class '%s' from '%s'), task "
                                     "has no storage assigned" % (self.name, self.class_name, self.import_path))

        if 'selective_run_function' in opts:
            self.selective_run_function = SelectiveRunFunction.from_dict(opts.pop('selective_run_function'))
        else:
            self.selective_run_function = SelectiveRunFunction.get_default()

        self.storage_task_name = opts.pop('storage_task_name', name)
        self.output_schema = opts.pop('output_schema', None)

        if opts.get('retry_countdown') is not None and opts.get('max_retry', 0) == 0:
            self._logger.warning("Retry countdown set for task '%s' (class '%s' from '%s') but this task has"
                                 "retry set to 0", self.name, self.class_name, self.import_path)

        self.max_retry = opts.pop('max_retry', self._DEFAULT_MAX_RETRY)
        self.retry_countdown = opts.pop('retry_countdown', self._DEFAULT_RETRY_COUNTDOWN)

        self.queue_name = self._expand_queue_name(opts.pop('queue', None))
        self.storage_readonly = opts.pop('storage_readonly', False)
        self.throttling = self.parse_throttling(opts.pop('throttling', {}))

        if opts:
            raise ConfigurationError("Unknown task option provided for task '%s' (class '%s' from '%s'): %s"
                                     % (name, self.class_name, self.import_path, opts))

        # register task usage
        if self.storage:
            storage.register_task(self)

        self._logger.debug("Creating task with name '%s' import path '%s', class name '%s'",
                           self.name, self.import_path, self.class_name)

    def check(self):
        """Check task definitions for errors.

        :raises: ValueError if an error occurred
        """
        if not isinstance(self.import_path, str):
            raise ConfigurationError("Error in task '%s' definition - import path should be string; got '%s'"
                                     % (self.name, self.import_path))

        if self.class_name is not None and not isinstance(self.class_name, str):
            raise ConfigurationError("Error in task '%s' definition - class instance should be string; got '%s'"
                                     % (self.name, self.class_name))

        if self.output_schema is not None and not isinstance(self.output_schema, str):
            raise ConfigurationError("Error in task '%s' definition - output schema should be string; got '%s'"
                                     % (self.name, self.output_schema))

        if self.max_retry is not None and (not isinstance(self.max_retry, int) or self.max_retry < 0):
            raise ConfigurationError("Error in task '%s' definition - max_retry should be None, zero or positive "
                                     "integer; got '%s'" % (self.name, self.max_retry))

        if self.retry_countdown is not None and (not isinstance(self.retry_countdown, int) or self.retry_countdown < 0):
            raise ConfigurationError("Error in task '%s' definition - retry_countdown should be None or positive "
                                     "integer; got '%s'" % (self.name, self.retry_countdown))

        if self.queue_name is not None and not isinstance(self.queue_name, str):
            raise ConfigurationError("Invalid task queue, should be string, got %s" % self.queue_name)

        if not isinstance(self.storage_readonly, bool):
            raise ConfigurationError("Storage usage flag readonly should be of type bool")

    @staticmethod
    def from_dict(dictionary, system):
        """Construct task from a dict and check task's definition correctness.

        :param dictionary: dictionary to be used to construct the task
        :type dictionary: dict
        :param system: system that should be used to for lookup a storage
        :type system: System
        :return: Task instance
        :rtype: selinon.selinon_task.Task
        :raises: ValueError
        """
        if 'name' not in dictionary or not dictionary['name']:
            raise ConfigurationError('Task name definition is mandatory')
        if 'import' not in dictionary or not dictionary['import']:
            raise ConfigurationError('Task import definition is mandatory')
        if 'storage' in dictionary:
            storage = system.storage_by_name(dictionary.pop('storage'))
        else:
            storage = None

        instance = Task(dictionary.pop('name'), dictionary.pop('import'), storage, **dictionary)
        instance.check()
        return instance



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""PostgreSQL models definition."""

try:
    from sqlalchemy import Column
    from sqlalchemy import Integer
    from sqlalchemy import Sequence
    from sqlalchemy import String
    from sqlalchemy.dialects.postgresql import JSONB
    from sqlalchemy.ext.declarative import declarative_base
except ImportError as exc:
    raise ImportError("Please install dependencies using `pip3 install selinon[postgresql]`") from exc

_Base = declarative_base()  # pylint: disable=invalid-name


class Result(_Base):
    """Record for a task result."""

    __tablename__ = 'result'

    id = Column(Integer, Sequence('result_id'), primary_key=True)  # pylint: disable=invalid-name
    flow_name = Column(String(128))
    task_name = Column(String(128))
    task_id = Column(String(255), unique=True)
    # We are using JSONB for postgres, if you want to use other database, change column type
    result = Column(JSONB)
    node_args = Column(JSONB)

    def __init__(self, node_args, flow_name, task_name, task_id, result):  # noqa
        self.flow_name = flow_name
        self.task_name = task_name
        self.task_id = task_id
        self.result = result
        self.node_args = node_args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import os
import pytest
from flexmock import flexmock
from jsonschema import ValidationError
from selinon_test_case import SelinonTestCase
from selinon.task_envelope import SelinonTaskEnvelope
from request_mock import RequestMock


class TestSelinonTaskEnvelope(SelinonTestCase):
    def test_validate_schema(self):
        edge_table = {}
        task_name = 'Task1'
        result = {'projectName': 'Selinon'}
        schema_path = os.path.join('test', 'data', 'validate_schema.json')

        self.init(edge_table, output_schemas={'Task1': schema_path})

        # we shouldn't raise anything
        SelinonTaskEnvelope.validate_result(task_name, result)

    def test_validate_schema_error(self):
        edge_table = {}
        task_name = 'Task1'
        result = {'foo': 'bar'}
        schema_path = os.path.join('test', 'data', 'validate_schema.json')

        self.init(edge_table, output_schemas={'Task1': schema_path})

        with pytest.raises(ValidationError):
            SelinonTaskEnvelope.validate_result(task_name, result)

    def test_selinon_retry(self):
        task = SelinonTaskEnvelope()
        task.request = RequestMock()
        flexmock(task).should_receive('retry').and_return(ValueError)

        params = {
            'task_name': 'Task1',
            'flow_name': 'flow1',
            'parent': {},
            'node_args': None,
            'retry_countdown': 0,
            'retried_count': 0,
            'dispatcher_id': '<dispatcher-id>',
            'user_retry': False
        }

        try:
            # we need to make context for traceback, otherwise this test will fail on Python3.4
            raise KeyError()
        except KeyError:
            # we should check params here
            with pytest.raises(ValueError):
                task.selinon_retry(**params)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from selinon.selective import compute_selective_run
from selinon_test_case import SelinonTestCase
from selinon import SelectiveNoPathError


class TestSelective(SelinonTestCase):
    @classmethod
    def _lists2sets(cls, d):
        for k, v in d.items():
            if isinstance(v, dict):
                d[k] = cls._lists2sets(v)
            elif isinstance(v, list) or isinstance(v, tuple):
                d[k] = set(v)
            else:
                d[k] = v
        return d

    def test_compute_selective_run_simple(self):
        #
        # flow1:
        #         |     |
        #        T1    T2
        #       / |     |
        #      /   -----
        #     /      |
        #   T4       T3
        #    |       | \
        #     -------   \
        #        |       T6
        #       T5
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true},
                      {'from': ['Task1', 'Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task3', 'Task4'], 'to': ['Task5'], 'condition': self.cond_true},
                      {'from': ['Task3'], 'to': ['Task6'], 'condition': self.cond_true}]
        }
        self.init(edge_table)
        selective = compute_selective_run('flow1', ['Task4', 'Task3'], follow_subflows=False, run_subsequent=False)
        self._lists2sets(selective)

        assert selective['waiting_edges_subset'] == {'flow1': {0: {'Task1', 'Task2'}, 1: {'Task3'}, 2: {'Task4'}}}
        assert set(selective['task_names']) == {'Task4', 'Task3'}

    def test_compute_selective_run_cyclic(self):
        #
        # flow1:
        #         |     |
        #        T1    T2 <-
        #         |     |   |
        #        T3    T4---
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)
        selective = compute_selective_run('flow1', ['Task4'], follow_subflows=False, run_subsequent=False)
        self._lists2sets(selective)

        assert selective['task_names'] == {'Task4'}
        # Yes, the third edge is correctly here as we expect all paths to Task4 to be present (even cyclic)
        assert selective['waiting_edges_subset'] == {'flow1': {0: {'Task2'}, 2: {'Task4'}, 3: {'Task2'}}}

    def test_compute_selective_run_multi_path(self):
        #
        # flow1:
        #         |      |
        #        T1     T2
        #         |      |
        #        T3     T3
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)
        selective = compute_selective_run('flow1', ['Task3'], follow_subflows=False, run_subsequent=False)
        self._lists2sets(selective)

        assert selective['task_names'] == {'Task3'}
        assert selective['waiting_edges_subset'] == {'flow1': {0: {'Task1'}, 1: {'Task2'}, 2: {'Task3'}, 3: {'Task3'}}}

    def test_compute_selective_run_follow_subflows(self):
        #
        # flow1:
        #
        #      |
        #     T1
        #      |\
        #     T2 f2
        #      |
        #     T3
        #
        #
        # flow2:
        #
        #      |
        #     T4
        #      |
        #     T2
        #      |
        #     T5
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true}],
            'flow2': [{'from': [], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task5'], 'condition': self.cond_true}]
        }
        self.init(edge_table)
        selective = compute_selective_run('flow1', ['Task2'], follow_subflows=True, run_subsequent=False)
        self._lists2sets(selective)

        assert selective['waiting_edges_subset'] == {'flow1': {0: {'Task1'}, 1: {'Task2'}, 2: {'flow2'}},
                                                     'flow2': {0: {'Task4'}, 1: {'Task2'}}}
        assert selective['task_names'] == {'Task2'}

    def test_compute_selective_run_follow_subflows_recursive(self):
        #
        # flow1:
        #
        #      |
        #     T1
        #      |\
        #     T2 f2
        #      |
        #     T3
        #
        #
        # flow2:
        #
        #      |
        #     T4
        #      |\
        #     T2 f3
        #      |
        #     T5
        #
        # flow3:
        #
        #      |
        #     T6
        #      |\
        #     T2 f1
        #      |
        #     T7
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true}],
            'flow2': [{'from': [], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['flow3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task5'], 'condition': self.cond_true}],
            'flow3': [{'from': [], 'to': ['Task6'], 'condition': self.cond_true},
                      {'from': ['Task6'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task6'], 'to': ['flow1'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task7'], 'condition': self.cond_true}]
        }
        self.init(edge_table)
        selective = compute_selective_run('flow1', ['Task2'], follow_subflows=True, run_subsequent=False)
        self._lists2sets(selective)

        assert selective['waiting_edges_subset'] == {'flow2': {0: {'Task4'}, 1: {'Task2'}, 2: {'flow3'}},
                                                     'flow1': {0: {'Task1'}, 1: {'Task2'}, 2: {'flow2'}},
                                                     'flow3': {0: {'Task6'}, 1: {'Task2'}}}
        assert selective['task_names'] == {'Task2'}

    def test_compute_selective_run_subsequent(self):
        #
        # flow1:
        #
        #      |
        #     T1
        #      |
        #     T2
        #      |
        #     T3
        #
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true}],
        }
        self.init(edge_table)
        selective = compute_selective_run('flow1', ['Task2'], follow_subflows=False, run_subsequent=True)
        self._lists2sets(selective)

        assert selective['waiting_edges_subset'] == {'flow1': {0: {'Task1'}, 1: {'Task2'}, 2: {'Task3'}}}
        assert selective['task_names'] == {'Task2'}

    def test_compute_selective_run_follow_subflows_subsequent(self):
        #
        # flow1:
        #
        #      |
        #     T1
        #      |\
        #     T2 f2
        #      |
        #     T3
        #
        #
        # flow2:
        #
        #      |
        #     T4
        #      |
        #     T2
        #      |
        #     T5
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true}],
            'flow2': [{'from': [], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task5'], 'condition': self.cond_true}]
        }
        self.init(edge_table)
        selective = compute_selective_run('flow1', ['Task2'], follow_subflows=True, run_subsequent=True)
        self._lists2sets(selective)

        assert selective['waiting_edges_subset'] == {'flow1': {0: {'Task1'}, 1: {'Task2'}, 2: {'flow2'}, 3: {'Task3'}},
                                                     'flow2': {0: {'Task4'}, 1: {'Task2'}, 2: {'Task5'}}}
        assert selective['task_names'] == {'Task2'}

    def test_compute_selective_run_bad_task_names(self):
        #
        # flow1:
        #
        #      |
        #     T1
        #      |
        #     T2
        #      |
        #     T3
        #
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true}],
        }
        self.init(edge_table)
        with pytest.raises(SelectiveNoPathError):
            compute_selective_run('flow1', ['TaskX'], follow_subflows=False, run_subsequent=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""A queue that respect timestamps of records that were pushed into it."""

import heapq


class TimeQueue:
    """A queue that respect timestamps of records that were pushed into it."""

    class _TimeQueueItem:
        """TimeQueue internal item."""

        def __init__(self, time, record):
            """Instantiate item for TimeQueue.

            :param time: timestamp of record
            :param record: record that should be stored
            """
            self.time = time
            self.record = record

        def __lt__(self, other):  # noqa
            return self.time < other.time

        def __repr__(self):
            """Representation for nice logs/debug.

            :return: item representation
            """
            return "%s(%s, %s)" % (self.__class__.__name__, self.time, self.record)

    def __init__(self):
        """Init time queue."""
        self._queue = []

    def push(self, time, record):
        """Push record with the given timestamp to queue.

        :param time: time of the record
        :param record: record to be pushed
        """
        heapq.heappush(self._queue, self._TimeQueueItem(time, record))

    def pop(self):
        """Remove and return the top record in the queue.

        :return: time and record tuple that were stored
        """
        result = heapq.heappop(self._queue)
        return result.time, result.record

    def top(self):
        """Return the top record in the queue, do not remove it from the queue.

        :return: time and record tuple that were stored
        """
        result = self._queue[0]
        return result.time, result.record

    def is_empty(self):
        """Check queue emptiness.

        :return: True if queue is empty
        """
        return len(self._queue) == 0

    def __repr__(self):
        """Queue representation.

        :return: a string representation for this queue.
        """
        return repr(self._queue)
